'''A Qt-compatible interface similar to `chess.engine`_ 
 
.. _chess.engine: https://pypi.org/project/chess
'''

import os,  os.path
import time
import re
from typing import Any, Dict, Callable, Iterable, List, Optional, Tuple, Type, Union
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtCore
else:
 from PyQt6 import QtCore

import chess
import chess.engine

PGNCmd_REGEX = re.compile(r'\[(%[a-z]*)?[ ]+([^\n\t \]]+)\]')
PGNEval_REGEX = re.compile(r'\[(%|%eval)[ ]+([^\n\t \]]+)\]')

class ChessEngine(QtCore.QObject):
 '''A Universal Chess Interface (`UCI`_) engine using (`QProcess`_)

:param engine2Option: a pair of a path to the executable and *dict* of the changed options
:param limit: a limit definition for the engine (see `chess.engine.Limit`_)
:param timeout_msec: timeout for the engine to respond (should be at least 1000)
:param log: log for the engine's commands 
:param ChessEngine.bestMoveScoreSignal: ``pyqtSignal`` emitted when a best move is available
 
.. _UCI: http://wbec-ridderkerk.nl/html/UCIProtocol.html
.. _QProcess: https://doc.qt.io/qt-5/qprocess.html
.. _chess.engine.Limit: https://python-chess.readthedocs.io/en/latest/engine.html
 '''
 bestMoveScoreSignal : QtCore.pyqtSignal = QtCore.pyqtSignal(chess.Move, str)

 def __init__(self, engine2Option : Tuple[os.PathLike, Dict[str, Union[str, int, bool, None]]], 
   limit : chess.engine.Limit = chess.engine.Limit(depth = 10), 
   timeout_msec : int = 1000, 
   log : Optional[Callable[[str], None]] = None, 
   parent : Optional[QtCore.QObject] = None) -> None:
  super(ChessEngine, self).__init__(parent)
  executable, optionsDict = engine2Option
  if not (os.path.isfile(executable) and os.access(executable, os.X_OK )):
   raise IOError('ChessEngine: {} is not an executable'.format(executable))
  self.limit = limit
  self.setLog(log)
  self.timeout_msec = max(int(timeout_msec), 10)
  self.stdoutLines = list()
  self.p = QtCore.QProcess()  # Keep a reference to the QProcess (e.g. on self) while it's running.
  self.p.readyReadStandardOutput.connect(self._fromStdout)
  self.p.readyReadStandardError.connect(self._fromStdout)
  self.stdout = ''
  self.stdoutLines = list()
  self.p.start(executable, [], QtCore.QIODevice.OpenModeFlag.ReadWrite | QtCore.QIODevice.OpenModeFlag.Text)
  while not self.p.waitForStarted(msecs = self.timeout_msec):
   self._log('ChessEngine: slow write, {} msec elapsed'.format(self.timeout_msec))
  self.readyok = True
  self._toStdin('uci')
  start = int(round(time.time() * 1000))
  loop = 1
  while 'idDict' not in vars(self) or 'optionsDict' not in vars(self):
   QtCore.QCoreApplication.processEvents()
   elapsed = int(round(time.time() * 1000)) - start
   if elapsed > 3*self.timeout_msec:
    self._log('ChessEngine: long wait for uci response, {}x looped, {} msec elapsed'.format(loop, elapsed))
    break
   loop += 1
  if self.isReady():
   for name, value in optionsDict.items():
    self.uciSetOption(name, value)
  self.isrunning = False
  self.playResult = None
  self.board = chess.Board()

 def kill(self, beSilent : bool = False) -> None:
  '''Kills the process, engine cannot be used anymore
  '''
  if 'p' in vars(self):
   self.p.kill()
   isFinished = self.p.waitForFinished(1000)
   if not (isFinished and beSilent):
    raise IOError('Cannot stop ChessEngine ...')
   self.p = None

 def setLog(self, log : Optional[Callable[[str], None]] = None) -> None:
  '''Sets the log for engine's commands

:param log: log for the engine's commands, default = None  
  '''
  if isinstance(log, bool):
   if log:
    self.notify = print
   else:
    self.notify = None
  else:
   self.notify = log

 def getScore(self, hintID : int = 0) -> Optional[int]:
  '''Delivers the score for a hint, if available

:param hintID: hint, i.e. alternative to be used (0 for best move)
:returns: score in centipawns or ``None``
  '''
  if   'info' not in vars(self.playResult) \
    or hintID >= len(self.playResult.info) \
    or 'score' not in self.playResult.info[hintID]:
   return None
  score = self.playResult.info[hintID]['score'].white()
  if isinstance(score, chess.engine.Cp):
   score = '{:+.1f}'.format(0.01 * score.score())
  elif score is not None:
   score = str(score)
  return score
  
 def _log(self, txt : str) -> None:
  if self.notify is not None:
   self.notify(txt)

 def _fromStdout(self) -> None:
  if self.p is None:
   return
  if 'stdout' not in vars(self):
   self.stdout = ''
  try:
   self.stdout += bytes(self.p.readAllStandardOutput()).decode('utf-8')
  except:
   return
  if not self.stdout.endswith('\n'):
   return
  self.stdoutLines += self.stdout.strip("\n").split("\n")
  self._log('ChessEngine/_fromStdout: stdout< {}'.format(self.stdout))
  self.stdout = ''
  if self.stdoutLines[-1] == 'uciok':
   self._parseHeader()
   self.readyok = True
   self.stdoutLines = list()
  elif self.stdoutLines[-1] == 'readyok':
   self.readyok = True
   self.stdoutLines = list()
  elif self.stdoutLines[-1].startswith('bestmove'):
   self._parseResult()
   score = self.getScore(hintID = 0)
   self.readyok = True
   self.stdoutLines = list()
   if len(self.playResult.info) == 1: 
    self.playResult.info = self.playResult.info[0]
   if self.playResult.move is not None:
    self.bestMoveScoreSignal.emit(self.playResult.move, score)
   elif score is not None:
    self.bestMoveScoreSignal.emit(chess.Move.null(), score)

 def _fromStderr(self) -> None:
  stderr = bytes(self.p.readAllStandardError()).decode('utf-8').strip("\n")
  self._log('ChessEngine/_fromStderr:???????????????\n <stderr: {} \n???????????????'.format(stderr))

 def _toStdin(self, txt : str) -> bool:
  if self.p is None:
   return False
  self._log('ChessEngine/_toStdin: stdin> {}, readyok = {}'.format(txt, self.readyok))
  if not self.readyok:
   return False
  self.readyok = False
  self.stdoutLines = list()
  self.p.write("{}\n".format(txt).encode('utf-8'))
  while not self.p.waitForBytesWritten(msecs = self.timeout_msec):
   self._log('ChessEngine/_toStdin: slow write, {} msec elapsed'.format(self.timeout_msec))
  return True
 
 def isReady(self) -> bool:
  '''Checks whether the engine is able to receive commands

:returns: boolean indicating the response
  '''
  return self.p.state() == QtCore.QProcess.ProcessState.Running and self.readyok
  
 #  Parser --------------------------------------------------------------------------------------

 def _parseHeader(self) -> None:
  self.idDict = dict()
  self.optionsDict = dict()
  for line in self.stdoutLines:
   tokens = line.split(" ")
   report = tokens.pop(0)
   if report == 'id':
    key = tokens.pop(0)
    if key != 'name' and key != 'author':
     raise ValueError("UIE: Unknown {} in header (expected name or author)".format(key))
    self.idDict[key] = ' '.join(tokens)
   elif report == 'option':
    self._parseOption(tokens)
   elif len(report.strip(' ')) > 0 and report != 'uciok':
    self._log("Extra line '{}'".format(line))

 def _parseOption(self, tokens) -> None:
  optDict = dict()
  minValue = None 
  maxValue = None
  name = None
  vtype = None
  while tokens:
   key = tokens.pop(0)
   if key == 'name':
    nameList = list()
    while tokens:
     part = tokens[0]
     if part in  ['type', 'default', 'min', 'max', 'var']:
      break
     tokens.pop(0)
     nameList.append(part)
    name = ' '.join(nameList)
   elif key == 'type':
    vtype = tokens.pop(0)
    if vtype not in ['check', 'spin', 'combo', 'button', 'string']:
     raise ValueError("UIE: Unexpected option type '{}'".format(vtype))
    if vtype == 'combo':
     optDict['varList'] = list()
    optDict[key] = vtype
   elif key == 'default':
    value = tokens.pop(0)
    if vtype == 'spin':
     optDict[key] = int(value)
    elif value == 'check':
     optDict[key] = value == 'true'
    else:
     optDict[key] = value
   elif key == 'min':
    minValue =  int(tokens.pop(0)) 
   elif key == 'max':
    maxValue =  int(tokens.pop(0)) 
   elif key == 'var':
    if 'varList' not in optDict:
     raise ValueError("UIE: var found, not type 'combo' ?!?")
    optDict['varList'].append(tokens.pop(0))
   elif key == 'string':
    if len(tokens) > 0:
     value = tokens.pop(0)
    else:
     value = ''
    if value == '<empty>':
     value = ''
    optDict[key] = value
  if name is None or vtype is None:
   raise ValueError("UIE: name and/or type record missing")
  if vtype == 'spin' and minValue is not None and maxValue is not None:
   optDict['range'] = range(minValue, maxValue+1)
  self.optionsDict[name] = optDict

 def _parseResult(self) -> None:
  depth = 0
  if self.playResult is None or 'move' in vars(self.playResult):
   self.playResult = chess.engine.PlayResult(None, None, draw_offered = None, resigned = None)
   self.playResult.info = list()
  for line in reversed(self.stdoutLines):
   tokens = line.split(" ")
   report = tokens.pop(0)
   if report == 'info':
    actInfo, newDepth = self._parseInfoline(tokens, depth)
    if newDepth is None or len(actInfo) == 0 or 'score' not in actInfo:
     continue
    if newDepth < depth:
     self.playResult.info = list(reversed(self.playResult.info))
     break
    depth = newDepth
    self.playResult.info.append(actInfo)
   elif report == 'bestmove':
    try:
     self.playResult.move = chess.Move.from_uci(tokens.pop(0))
    except:
     self.playResult.move = None
    if len(tokens) > 0 and tokens.pop(0) == 'ponder':
     try:
      self.playResult.ponder = chess.Move.from_uci(tokens.pop(0))
     except:
      self.playResult.ponder = None 

 def _parseInfoline(self, tokens, depth) -> Tuple[Dict[str, Any], int]:
   info = dict()
   actDepth = None
   while tokens:
    parameter = tokens.pop(0)
    if parameter == "string":
     info["string"] = " ".join(tokens)
     return dict(), depth
    elif parameter == "depth":
     actDepth = int(tokens.pop(0))
     info[parameter] = actDepth
    elif parameter in ["seldepth", "nodes", "multipv", "currmovenumber", "hashfull", "nps", "tbhits", "cpuload"]:
     info[parameter] = int(tokens.pop(0))  # type: ignore
    elif parameter == "time":
     info["time"] = int(tokens.pop(0)) // 1000.0
    elif parameter == "ebf":
     info["ebf"] = float(tokens.pop(0))
    elif parameter == "score":
     kind = tokens.pop(0)
     value = tokens.pop(0)
     if tokens and tokens[0] in ["lowerbound", "upperbound"]:
      info[tokens.pop(0)] = True  # type: ignore
     if kind == "cp":
      info["score"] = chess.engine.PovScore(chess.engine.Cp(int(value)), self.board.turn)
     elif kind == "mate":
      info["score"] = chess.engine.PovScore(chess.engine.Mate(int(value)), self.board.turn)
     else:
      raise ValueError("UIE: Unknown score kind {} in info (expected cp or mate)".format(kind))
    elif parameter == "currmove":
     info["currmove"] = chess.Move.from_uci(tokens.pop(0))
    elif parameter == "currline":
     cpunr = int(tokens.pop(0))
     currline = list()
     for n in range(cpunr):
      currline.append(tokens.pop(0))
     info["currline"] = currline
    elif parameter == "refutation":
     info["refutation"] = []
     while tokens:
      try:
       move = chess.Move.from_uci(tokens[0])
       tokens.pop(0)
       info["refutation"].append(move)
      except:
       break
    elif parameter == "pv":
     info["pv"] = list()
     while tokens:
      try:
       move = chess.Move.from_uci(tokens[0])
       tokens.pop(0)
       info["pv"].append(move)
      except:
       break
    elif parameter == "wdl":
     info["wdl"] = chess.engine.PovWdl(chess.engine.Wdl(int(tokens.pop(0)), int(tokens.pop(0)), int(tokens.pop(0))), self.board.turn)
   return info, actDepth

 #  Methods for use  --------------------------------------------------------------------------------------

 def uciNewGame(self, fen : Optional[str] = None, moves : List[chess.Move] = []) -> bool:
  '''Wrapper for the UCI *ucinewgame* command

:param fen: starting position in Forsyth-Edwards-Notation (FEN)
:param moves: list of moves to be applied
:returns: boolean indicating the success
  '''
  self.playResult = None
  if not self._toStdin('ucinewgame'):
   return False
  builder = ["position"]
  if fen is not None:
   self.board.set_fen(fen)
   builder.append("fen")
   builder.append(fen)
  else:
   self.board.set_fen(chess.STARTING_FEN)
   builder.append("startpos")
  for move in moves:
   builder.append(move.uci())
  self.readyok = True
  self._toStdin(' '.join(builder))
  self.readyok = True
  return True

 def uciSetOption(self, name : str, value : Union[bool, int,  str, None]) -> bool:
  '''Wrapper for the UCI *setoption* command

:param name: name of the option
:param value: value, type depends on name
:returns: boolean indicating the success
  '''
  if name not in self.optionsDict:
   raise ValueError('ChessEngine/setOption: {} is not a valid option name'.format(name))
  optDict = self.optionsDict[name]
  optType = optDict['type']
  builder = ['setoption']
  builder.append('name')
  builder.append(name)
  builder.append('value')
  if optType == 'button' and value is not None:
    raise ValueError('ChessEngine/setOption: option name {} expects no value'.format(name))
  elif optType == 'check':
   if not isinstance(value, bool):
    raise ValueError('ChessEngine/setOption: option name {} expects an boolean value'.format(name))
   builder.append(['false', 'true'][value])
  elif optType == 'spin':
   if isinstance(value, str):
    value = int(value)
   if 'range' in optDict and value not in optDict['range']:
    raise ValueError('ChessEngine/setOption: option name {}: value {} not in range {}'.format(name, value, optDict['range']))
   builder.append(str(value))
  elif optType == 'combo':
   if not isinstance(value, str):
    raise ValueError('ChessEngine/setOption: option name {} expects an string value'.format(name))
   if 'varList' in optDict and value not in optDict['varList']:
    raise ValueError('ChessEngine/setOption: option name {}: value {} not in list {}'.format(name, value, optDict['varList']))
   builder.append(value)
  elif Type == 'string':
   if not isinstance(value, str):
    raise ValueError('ChessEngine/setOption: option name {} expects an string value'.format(name))
   if len(value) == 0:
    value == '<empty>'
   builder.append(value)
  if self._toStdin(' '.join(builder)):
   self.readyok = True
  return self.readyok

 def uciGo(self, search_moves : Optional[Iterable[chess.Move]] = None, ponder : bool = False, infinite : bool = False) -> bool:
  '''Wrapper for the UCI *go* command

:param search_moves: list of moves to be searched
:param ponder: suggest a response to the *best-move*
:param infinite: improve the *best-move* continuously
:returns: boolean indicating the success
  '''
  builder = ["go"]
  if ponder:
   builder.append("ponder")
  if self.limit.white_clock is not None:
   builder.append("wtime")
   builder.append(str(max(1, int(self.limit.white_clock * 1000))))
  if self.limit.black_clock is not None:
   builder.append("btime")
   builder.append(str(max(1, int(self.limit.black_clock * 1000))))
  if self.limit.white_inc is not None:
   builder.append("winc")
   builder.append(str(int(self.limit.white_inc * 1000)))
  if self.limit.black_inc is not None:
   builder.append("binc")
   builder.append(str(int(self.limit.black_inc * 1000)))
  if self.limit.remaining_moves is not None and int(self.limit.remaining_moves) > 0:
   builder.append("movestogo")
   builder.append(str(int(self.limit.remaining_moves)))
  if self.limit.depth is not None:
   builder.append("depth")
   builder.append(str(max(1, int(self.limit.depth))))
  if self.limit.nodes is not None:
   builder.append("nodes")
   builder.append(str(max(1, int(self.limit.nodes))))
  if self.limit.mate is not None:
   builder.append("mate")
   builder.append(str(max(1, int(self.limit.mate))))
  if self.limit.time is not None:
   builder.append("movetime")
   builder.append(str(max(1, int(self.limit.time * 1000))))
  if infinite:
   builder.append("infinite")
  if search_moves is not None:
   builder.append("searchmoves")
   builder.extend(move.uci() for move in search_moves)
  if not self._toStdin(' '.join(builder)):
   return False

  if infinite:
   self.isrunning = True
   self.readyok = True
  return True

 def uciStop(self) -> bool:
  '''Wrapper for the UCI *stop* command to stop an infinite *go*

:returns: boolean indicating the success
  '''
  if not self.isrunning:
   return False
  return self._toStdin('stop')

 def uciQuit(self) -> None:
  'Terminate the engine'
  self._toStdin('quit')

 #  chess.engine like calls  --------------------------------------------------------------------------------------

 def startPlay(self) -> bool:
  '''Emits *uciGO* in *play* mode, i.e.g
  
  * *UCI_AnalyseMode = off*
  * *MultiPV = 1*, i.e. no alternative move suggestions

:returns: boolean indicating the success
  '''
  if 'MultiPV' in self.optionsDict:
   if not self.uciSetOption('MultiPV', 1):
    return False
  if 'UCI_AnalyseMode' in self.optionsDict:
   if not self.uciSetOption('UCI_AnalyseMode', False):
    return False
  return self.uciGo()

 def startAnalysis(self, multiPV : int = 1) -> bool:
  '''Emits *uciGO* in *analyse* mode, i.e.
  
  * *UCI_AnalyseMode = off*
  * *MultiPV =* ``multiPV``

:param multiPV: number of alternative move suggestions
:returns: boolean indicating the success
  '''
  if 'MultiPV' in self.optionsDict:
   if not self.uciSetOption('MultiPV', max(1, multiPV)):
    return False
  if 'UCI_AnalyseMode' in self.optionsDict:
   if not self.uciSetOption('UCI_AnalyseMode', True):
    return False
  return self.uciGo()

 def setELO(self, elo : Union[int, str]) -> bool:
  if not ('UCI_Elo' in self.optionsDict and 'UCI_LimitStrength' in self.optionsDict):
   return False
  eloDict = self.optionsDict['UCI_Elo']
  if elo == 'max':
   elo =eloDict['range'].stop - 1
  elif elo == 'min':
   elo = eloDict['range'].start
  if elo not in eloDict['range']:
   raise ValueError('ChessEngine: elo {} not in {}'.format(elo, eloDict['range']))
  if not self.uciSetOption('UCI_Elo', elo):
   return False
  if not self.uciSetOption('UCI_LimitStrength', True):
   return False
  self.readyok = True
  return True
  
# ---------------------------------------------------------------------
# ---------------------------------------------------------------------

if __name__ == "__main__":
 
 os.chdir('C:/Users/Reinh/chess_engines')
 stockfish12 = 'stockfish_12_win_x64_bmi2/stockfish_20090216_x64_bmi2.exe'
 togaII4 = 'TogaII40/Windows/TogaII_40_intelPGO_x64.exe'
 komodo12 = 'komodo-12.1.1_5a8fc2/Windows/komodo-12.1.1-64bit-bmi2.exe'
 fritz17 = 'Fritz17/Fritz17.exe'
 lcZero26 = 'lc0-v0.26.3-windows-gpu-opencl/lc0.exe'
 raubfischGTZ23 = 'RaubfischX44_and_GTZ23/GTZ23/RaubfischGTZ23_nn_sl-bmi2.exe'
 
 executable = stockfish12

 def waitForData(engine):
  for i in range(100):
   QtCore.QThread.msleep(engine.timeout_msec)
   QtCore.QCoreApplication.processEvents()
   if engine.isReady():
    return
   print('----> waiting {} * {} ms'.format(i, engine.timeout_msec))

 def printPlayResult(engine):
  playResult = engine.playResult
  rootBoard = chess.Board(engine.board.fen())
  print('{} : {}'.format('best move', playResult.move))
  if playResult.ponder is not None:
   print('{} : {}'.format('expected response', playResult.ponder))
  if isinstance(playResult.info, dict):
   infoList = [playResult.info]
  else:
   infoList = playResult.info
  for i, infoDict in enumerate(infoList):
   print('info #{}'.format(i+1))
   for key, value in infoDict.items(): 
    if key == 'pv':
     pvString = ''
     for n, moveList in enumerate(value):
      if not isinstance(moveList, list):
       moveList = [moveList]
      for move in moveList: 
       rootBoard.push(move)
       uciMove = move.uci()
       if rootBoard.is_checkmate():
        uciMove += '#'
       elif rootBoard.is_check():
        uciMove += '+'
       if engine.board.turn:
        if n == 0:
         pvString += '1.{}'.format(uciMove)
        elif n % 2 == 0:
         pvString += ' {}.{}'.format(n//2+1, uciMove)
        else:
         pvString += ' {}'.format(uciMove)
       else:
        if n == 0:
         pvString += '1...{}'.format(uciMove)
        elif n % 2 == 1:
         pvString += ' {}.{}'.format((n+1)//2+1, uciMove)
        else:
         pvString += ' {}'.format(uciMove)
       print('{} : {}'.format(key, pvString))
    elif value is not None: 
     print('{} : {}'.format(key, value))

 def runChessEngine():
  global executable
  fenList = [
  "3rr3/2p3p1/4N3/3b4/1p3P2/2PBB3/kPQ3PP/3R2K1 w - - 0 1", 
  "rn3r1k/pp4pp/2p1Q2N/q2np3/4N2P/2PP4/PP3P2/R3K2R w KQ - 0 1", 
  "r2rkn2/1R1N2p1/2p1B2p/4p1PP/p1P2b2/5P2/P3K3/1R6 w - - 0 1", 
  "2r4k/p2R4/1p2P3/5p2/3PbbpP/BP6/P1r5/4Q1K1 b - - 0 1"
  ]
  engine = ChessEngine(executable, limit = chess.engine.Limit(depth = 15),  log = True)
  waitForData(engine)
  for item in sorted(engine.idDict):
   print(' {} : {}'.format(item, engine.idDict[item]))
  print('Options:')
  for opt in sorted(engine.optionsDict):
   print(' {} : {}'.format(opt, engine.optionsDict[opt]))
  for fen in fenList:
   engine.uciNewGame(fen = fen)
   print('------------------------- startPlay -------------------------')
   engine.startPlay()
   waitForData(engine)
   printPlayResult(engine)
   print('----------------------- startAnalysis -----------------------')
   engine.startAnalysis(multiPV = 1)
   waitForData(engine)
   printPlayResult(engine)

 import sys
 from eco import ECODatabase 
 def scoreECOTable():
  global executable, basename, depth
  filename = '{}.tsv'.format(basename)
  print('Scoring {} -------------'.format(filename))
  eco = ECODatabase(tsvPattern = filename)
  fenList = eco.fen2Id().keys()
  engine = ChessEngine(executable, 
   limit = chess.engine.Limit(depth = depth),  
   timeout_msec = 300, 
   log = True)
  waitForData(engine)
  nTotal = len(fenList)
  fen2score = ['\t'.join(['fen',executable])]
  scmin, scmean, scmax = sys.maxsize//2, 0, -(sys.maxsize//2)
  for n, fen in enumerate(fenList):
   engine.startAnalysis(multiPV = 1)
   waitForData(engine)
   score = int(str(engine.playResult.info['score'].relative))
   scmin = min(scmin, score)
   scmean += score/nTotal
   scmax = max(scmax, score)
   fen2score.append('\t'.join([fen, str(score)]))
   print(' {} of {} completed'.format(n+1, nTotal))
  print('Scores {} <= {:.1f} <= {}'.format(scmin, scmean, scmax))

  fscFile = os.path.join(eco.ecoDirectory, '{}.fsc'.format(basename))
  with open(fscFile, mode='w') as f:
   f.write('\n'.join(fen2score))
  print('Result filed @ {}'.format(fscFile))

 app = QtCore.QCoreApplication(sys.argv)

 QtCore.QTimer.singleShot(10, runChessEngine)
 # QtCore.QTimer.singleShot(10, scoreECOTable)
 
 app.exec()
 print('completed')



from sys import version_info, path
import os.path

path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets, QtCore
 from PyQt5 import uic
else:
 from PyQt6 import QtWidgets, QtCore
 from PyQt6 import uic

import chess
import ply

class AboutDialog(QtWidgets.QDialog):
 """
 Class documentation goes here.
 """
 def __init__(self, parent=None):
  """
  Constructor
  
  @param parent reference to the parent widget
  @type QWidget
  """
  super(AboutDialog, self).__init__(parent)
  fileDirectory = os.path.dirname(os.path.abspath(__file__))
  uic.loadUi(os.path.join(fileDirectory, 'AboutDialog.ui'), self)

 def setup(self, 
  pgm = None, 
  version = '?.?.?', 
  dateString = 'Jan 01, 1970'):
  self.pgmLabel.setText(pgm)
  self.shoeboxVersionLabel.setText('V' + version)
  self.shoeboxDateLabel.setText(dateString)
  self.pythonVersionLabel.setText('V' + str(version_info[0]) + '.' + str(version_info[1]) + '.' + str(version_info[2]) ) 
  self.qtVersionLabel.setText('V' + QtCore.QT_VERSION_STR ) 
  self.pyqt6VersionLabel.setText('V' + QtCore.PYQT_VERSION_STR )
  self.chessVersionLabel.setText('V' + chess.__version__)
  self.plyVersionLabel.setText('V' + ply.__version__)

if __name__ == "__main__":
 import sys

 app =  QtWidgets.QApplication(sys.argv)
 dialog = AboutDialog()
 dialog.setup('About')
 dialog.show()
 sys.exit(app.exec())



'''
Score Chart
=============

|ScoreChart|

The score chart shows the material score (blue) representing material budget where 
the pieces represent as usual the values

* :math:`s_{pawn} = 1`
* :math:`s_{knight} = 3`
* :math:`s_{bishop} = 3`
* :math:`s_{rook} = 5`
* :math:`s_{queen} = 9`

If available, the scores emitted by a chess engine (red) are also shown.
The engine annotations require command tags according to `PGNExt`_ supplement, i.e. tags like [%eval *score*] or [%  *score*]
*score* is the score in pawns. 

The handling of positions close to mate vary from GUI to GUI. This GUI delivers for
a mate in *n* moves a score of

 * *mateScore - n*, if black is facing a mate in $n$
 * *n -mateScore*, if white is facing a mate in $n$

with :math:`mateScore = 100`.

The actual position of the game and the corresponding score values are indicated in the score graph.
It is updated when it changes.

It is possible to control the zoom of the *Score* axis. The GUI with an ongoing zoom process is shown here.

.. csv-table:: Zoom control by mouse
   :header: "Key", "Description"
   :widths: 30, 50

   :kbd:`mouse-left-press`       , begin zoom
   :kbd:`mouse-left-release`     , end zoom
   :kbd:`mouse-right-release`   , reset zoom

.. |ScoreChart| image:: scoreplotgraphicsview.png
  :width: 800
  :alt: Score Chart
.. _PGNExt: https://github.com/mliebelt/pgn-spec-commented/blob/main/pgn-spec-supplement.md
'''

from typing import Optional

import math
import sys, os, os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets, QtGui, QtCore
 from PyQt5 import QtChart as QtCharts
 from PyQt5.QtWidgets import QShortcut
else:
 from PyQt6 import QtCharts, QtWidgets, QtGui, QtCore
 from PyQt6.QtGui import QShortcut

import chess, chess.pgn
from chessengine import PGNEval_REGEX

class ScorePlot(QtCharts.QChartView):
 '''Score plot object
 '''
 piecePawnScoreDict = {
  chess.PAWN : 1, 
  chess.KNIGHT : 3, 
  chess.BISHOP : 3, 
  chess.ROOK : 5, 
  chess.QUEEN : 9
 }
 seriesPens = {
  'Material' : QtGui.QPen(QtGui.QBrush(QtCore.Qt.GlobalColor.blue), 1), 
  'Engine' : QtGui.QPen(QtGui.QBrush(QtCore.Qt.GlobalColor.red), 1), 
  None : QtGui.QPen(QtGui.QBrush(QtCore.Qt.GlobalColor.gray), 2, QtCore.Qt.PenStyle.DotLine)}
  
 axesPen = QtGui.QPen(QtGui.QBrush(QtCore.Qt.GlobalColor.blue), 2)
 axesBrush = QtGui.QBrush(QtCore.Qt.GlobalColor.black)
 mateScore = 100

 QtGui.QColor(QtCore.Qt.GlobalColor.blue)

 def __init__(self, parent : Optional[QtCore.QObject] = None) -> None:
  super(ScorePlot, self).__init__(parent)
  self.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
  self.setRubberBand(QtCharts.QChartView.RubberBand.VerticalRubberBand)
 
 def setup(self, notifyGameNodeSelectedSignal : Optional[QtCore.pyqtSignal] = None):
  '''Set up of the score
  
:param notifyGameNodeSelectedSignal: signal to be emitted if a game node is selected
  '''
  if 'xAxis' in vars(self):
   return
  self.notifyGameNodeSelectedSignal = notifyGameNodeSelectedSignal 
  if self.notifyGameNodeSelectedSignal is not None:
   scUp = QShortcut(self)
   scUp.setKey(QtCore.Qt.Key.Key_Up)
   scUp.activated.connect(self.on_sc_activated)
   scDown = QShortcut(self)
   scDown.setKey(QtCore.Qt.Key.Key_Down)
   scDown.activated.connect(self.on_sc_activated)
 
 def _setupChart(self, notifyGameNodeSelectedSignal : Optional[QtCore.pyqtSignal] = None):
  self._createChart()
  self.xAxis = self._addAxis(QtCore.Qt.AlignmentFlag.AlignBottom, title = 'Move')
  self.yAxis = self._addAxis(QtCore.Qt.AlignmentFlag.AlignLeft, title = 'Score [pawns]')
  self.materialSeries = self._addSeries('Material', isLineSeries = True)
  self.engineSeries = self._addSeries('Engine', isLineSeries = True)
  self.selectedGameNode = None
  self.vLine = self._addSeries(None, isLineSeries = True)
  self.vLine.setPointsVisible(True)
  self.vLine.setPointLabelsVisible(True)
  self.vLine.setPointLabelsClipping(False)
  self.vLine.setPointLabelsFormat('@xPoint')
  vLineMarkers = self.chart.legend().markers(self.vLine)
  vLineMarkers[0].setVisible(False)
  self.meLabels = self._addSeries(None, isLineSeries = False)
  self.meLabels.setMarkerSize(10)
  self.meLabels.setPointsVisible(True)
  self.meLabels.setPointLabelsVisible(True)
  self.meLabels.setPointLabelsClipping(False)
  self.meLabels.setPointLabelsFormat('@yPoint')
  meLabelMarkers = self.chart.legend().markers(self.meLabels)
  meLabelMarkers[0].setVisible(False)

 @staticmethod
 def minQtVersion(major : int, minor : int, patch : int = 0) -> bool:
  '''Checks whether the actual Qt-version is smaller or equal than the specified version
  
:param major: specified major version number
:param minor: specified minor version number
:param patch: specified patch number
:returns: True, if the actual Qt-version is smaller or equal than the specified version
  '''
  qtMajor, qtMinor, qtPatch = list(map(int, QtCore.QT_VERSION_STR.split('.')))
  return not (qtMajor < major \
              or (qtMajor == major and qtMinor < minor) \
              or (qtMajor == major and qtMinor == minor and qtPatch <= patch))
              
 def interval(self, minV : float, maxV : float) -> int:
  assert maxV > minV
  if maxV - minV < 10:
   return 1
  if maxV - minV < 100:
   return 10
  return 100
  
 def _createChart(self, 
   bColor = QtCore.Qt.GlobalColor.lightGray, 
   paColor = QtCore.Qt.GlobalColor.white, 
   legendPosition = QtCore.Qt.AlignmentFlag.AlignBottom) -> None:
  self.chart = QtCharts.QChart()
  self.chart.setBackgroundBrush(bColor)
  self.chart.setPlotAreaBackgroundBrush(paColor)
  self.chart.setPlotAreaBackgroundVisible(True)
  self.chart.legend().setVisible(True)
  self.chart.legend().setAlignment(legendPosition)
  self.setChart(self.chart)

 def _addSeries(self, name : Optional[str] = None, isLineSeries : bool = True) -> QtCharts.QLineSeries:
  if isLineSeries:
   series = QtCharts.QLineSeries(self.chart)
   if name is not None:
    series.setName(name)
   series.setPen(self.seriesPens[name])
  else:
   series = QtCharts.QScatterSeries(self.chart)
   series.setPen(self.axesPen)
  self.chart.addSeries(series)
  series.attachAxis(self.xAxis)
  series.attachAxis(self.yAxis)
  return series

 def _addTextLabel(self) -> QtWidgets.QGraphicsSimpleTextItem:
  label = self.scene().addSimpleText('')
  label.hide()
  return label

 def _addAxis(self, alignment : QtCore.Qt.AlignmentFlag, title: str = '') -> QtCharts.QValueAxis:
  axis = QtCharts.QValueAxis()
  axis.setLinePen(self.axesPen)
  axis.setLabelsBrush(self.axesBrush)
  axis.setGridLineVisible(True)
  axis.setMinorTickCount(5)
  if self.minQtVersion(5, 12):
   axis.setTickType(QtCharts.QValueAxis.TickType.TicksDynamic)
   axis.setTickAnchor(0)
  if len(title) > 0: 
   axis.setTitleText(title)
  self.chart.addAxis(axis, alignment)
  return axis
  
 def _setRange(self, axis : QtCharts.QValueAxis, minV : float, maxV : float ) -> None:
  assert maxV > minV, '{} > {} required'.format(maxV, minV)
  if maxV - minV < 5:
   delta = 1
  elif maxV - minV < 10:
   delta = 1
  elif maxV - minV < 50:
   delta = 5
  elif maxV - minV < 100:
   delta = 10
  else:
   delta = 100
  axis.setRange(minV, maxV)
  if self.minQtVersion(5, 12):
   axis.setTickInterval(delta)
  else:
   minV = (minV // delta) * delta
   maxV = (maxV // delta + 1) * delta
   axis.setMin(minV)
   axis.setMax(maxV)
   axis.setTickCount(int((maxV-minV) // delta) + 1)
   
 def _move(self, ply : int) -> float:
  return (ply + 1) / 2

 def resetChart(self) -> None:
  '''Clears the chart
  '''
  self.engineDict = dict()
  self.minY = float('inf')
  self.maxY = -float('inf')
  self.xAxis.setRange(0, 1)
  self.yAxis.setRange(-1, 1)
  self.materialSeries.clear()
  self.engineSeries.clear()
  self.vLine.clear()
  self.meLabels.clear()
  return

 def copyAsBitmap(self, width : int = None) -> None:
  '''Copies the chart as a bitmap to clipboard
  
:param width: width of the bitmap (Def: current width of the bounding box)
  '''
  sourceRect = self.sceneBoundingRect().toRect()
  if width is None:
   width = sourceRect.width()
  width = int(width)
  height = int(sourceRect.height() * width / sourceRect.width())
  bgColor = self.scene().views()[0].backgroundBrush().color().toRgb()
  plt = QtGui.QImage(width, height,  QtGui.QImage.Format.Format_RGB32)
  plt.fill(bgColor)
  painter = QtGui.QPainter()
  painter.begin(plt)
  painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
  self.render(painter)
  painter.end()
  clipboard = QtWidgets.QApplication.clipboard()
  clipboard.setImage(plt)

 def selectNodeItem(self, gameNode : chess.pgn.GameNode) -> None:
  '''Selects a game node
  
:param gameNode: game node to be selected
  '''
  if gameNode is None or gameNode.parent is None or not gameNode.is_mainline():
   return
  ply = gameNode.ply() - 1
  assert ply < self.materialSeries.count()
  self.selectedGameNode = gameNode
  self.vLine.clear()
  self.meLabels.clear()
  xValue = self.materialSeries.at(ply).x()
  self.vLine.append(xValue, self.yAxis.min())
  self.vLine.append(xValue, 1000)
  self.vLine.show()
  self.meLabels.append(self.materialSeries.at(ply))
  if ply in self.engineDict:
   ply = self.engineDict[ply]
   self.meLabels.append(self.engineSeries.at(ply))
  self.meLabels.show()

 def addGameNodes(self, gameNode :  chess.pgn.GameNode) -> None:
  '''Adds 1 or more nodes, parent node of first node must exist in the editor
  
:param gameNode: game node to be added (must be main_variation !!)
  '''
  if gameNode is None or not gameNode.is_mainline():
   return
  if 'materialSeries' not in vars(self):
   self._setupChart()
   self.resetChart()
  self.chart.show()
  ply = gameNode.ply()
  engineData = list()
  materialData = list()
  nMaterial = self.materialSeries.count()
  nEngine = self.engineSeries.count()
  while gameNode is not None:
   pieceMap = gameNode.board().piece_map()
   pawnScore = 0
   for piece in list(pieceMap.values()):
    if piece.piece_type != chess.KING:
     if piece.color == chess.WHITE:
      pawnScore += self.piecePawnScoreDict[piece.piece_type]
     else:
      pawnScore -= self.piecePawnScoreDict[piece.piece_type]
   materialData.append(QtCore.QPointF((ply +1)/2, pawnScore))
   self.minY = min(self.minY, pawnScore)
   self.maxY = max(self.maxY, pawnScore)
   match = PGNEval_REGEX.search(gameNode.comment)
   if match is not None and (match.group(1) == '%' or match.group(1) == '%eval'):
    engineScore = match.group(2)
    if engineScore != 'None':
     try:
      engineScore = float(engineScore)
     except:
      engineScore = float(engineScore[1:])
      engineScore = math.copysign(1,engineScore) * (self.mateScore - abs(engineScore))
     engineData.append(QtCore.QPointF((ply +1)/2, engineScore))
     self.engineDict[nMaterial] = nEngine
     self.minY = min(self.minY, engineScore)
     self.maxY = max(self.maxY, engineScore)
     nEngine += 1
   gameNode = gameNode.next()
   nMaterial += 1
   ply += 1

  if self.minY > self.maxY:
   return
   
  if len(materialData) > 0:
   self.materialSeries.append(materialData)
   if len(engineData) > 0:
    self.engineSeries.append(engineData)

  self._setRange(self.xAxis, 1, max(ply / 2, 2))
  if self.minY < self.maxY:
   self._setRange(self.yAxis, self.minY, self.maxY)
  else:
   self._setRange(self.yAxis, self.minY - 0.5, self.minY + 0.5)
  
  self.update()

 def removeLastNode(self) -> None:
  if self.materialSeries.count() == 0:
   return
  
  ply = self.materialSeries.count() - 1
  mMove = self.materialSeries.at(ply).x()
  self.materialSeries.remove(ply)
  ply += 2
  self.minY = float('inf')
  self.maxY = -float('inf')
  for n in range(self.materialSeries.count()):
   pawnScore = self.materialSeries.at(n).y()
   self.minY = min(self.minY, pawnScore)
   self.maxY = max(self.maxY, pawnScore)
  if self.engineSeries.count() > 0:
   lastID = self.engineSeries.count() - 1
   eMove = self.engineSeries.at(lastID).x()
   if mMove == eMove:
    self.engineSeries.remove(lastID)
   for n in range(self.engineSeries.count()):
    pawnScore = self.engineSeries.at(n).y()
    self.minY = min(self.minY, pawnScore)
    self.maxY = max(self.maxY, pawnScore)

  if self.minY > self.maxY:
   return

  self._setRange(self.xAxis, 1, max(ply / 2, 2))
  if self.minY != self.maxY:
   self._setRange(self.yAxis, self.minY, self.maxY)
  else:
   self._setRange(self.yAxis, self.minY - 0.5, self.minY + 0.5)

  self.update()
  
 def setGame(self, game : chess.pgn.Game) -> None:
  '''Sets a new game
  
:param game: game node to be set
  '''
  gameNode = game.next()
  if gameNode is None:
   if 'chart' in vars(self):
    self.chart.hide()
   return
  self.addGameNodes(gameNode)
  if 'Annotator' in game.headers:
   self.engineSeries.setName(game.headers['Annotator'])
  else:
   self.engineSeries.setName('Engine')
  self.selectNodeItem(gameNode)
 
 @QtCore.pyqtSlot()
 def on_sc_activated(self):
  if self.selectedGameNode is not None:
   sendingSC = self.sender()
   if sendingSC.key() == QtCore.Qt.Key.Key_Down:
    newGameNode = self.selectedGameNode.next()
   else:
    newGameNode = self.selectedGameNode.parent
   if newGameNode is not None:
    self.notifyGameNodeSelectedSignal.emit(newGameNode)

 @QtCore.pyqtSlot(QtGui.QMouseEvent)
 def mouseReleaseEvent(self, e):
  plies = self.materialSeries.count()
  if plies == 0:
   return
  super(ScorePlot, self).mouseReleaseEvent(e)
  if e.button() == QtCore.Qt.MouseButton.RightButton:
   self.chart.zoomReset()

  self._setRange(self.xAxis, 1, plies / 2)
  self._setRange(self.yAxis, self.yAxis.min(), self.yAxis.max())
  self.selectNodeItem(self.selectedGameNode)


if __name__ == "__main__":
 import io, sys
 from pgnParse import read_game


 class _My(QtWidgets.QMainWindow):
  def __init__(self, game):
   super().__init__()
   self.setWindowTitle('self.chart Formatting Demo')
   self.plot = ScorePlot(self)
   self.setCentralWidget(self.plot)
   self.resize(1200, 800)
  
  def setup(self):  
   self.plot.setup()
   self.plot.setGame(game)
   gameNode = game
   for i in range(10):
    gameNode = gameNode.next()
   self.plot.selectNodeItem(gameNode)
     
 
 ps = "C:/Users/Reinh/OneDrive/Dokumente/Schach/ps210105.pgn"
 with open(ps, mode = 'r',  encoding = 'utf-8') as f:
  newData = f.read()

 pgn = io.StringIO(newData)
 game = read_game(pgn)
 app = QtWidgets.QApplication([])
 plotWindow = _My(game)
 plotWindow.show()
 plotWindow.setup()
 sys.exit(app.exec())






import sys,  os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets
else:
 from PyQt6 import QtWidgets

class HelpBrowser(QtWidgets.QWidget):
 indexFile = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'doc_build', 'html', 'index.html')

 def __init__(self):
   super().__init__()
   
   self.tb = QtWidgets.QTextBrowser()
   self.tb.setAcceptRichText(True)
   self.tb.setOpenExternalLinks(True)
   self.tb.setOpenLinks(True)
   with open(self.indexFile, 'r',  encoding = 'utf-8-sig') as f:
    indexText = f.read()
   self.tb.setHtml(indexText)

   vbox = QtWidgets.QVBoxLayout()
   vbox.addWidget(self.tb, 1)

   self.setLayout(vbox)

   self.setWindowTitle('QTextBrowser')
   self.setGeometry(100, 100, 800, 800)
   self.show()


if __name__ == '__main__':
 if MzChess.useQt5():
  from PyQt5 import QtGui, QtCore
 else:
  from PyQt6 import QtGui, QtCore

 app = QtWidgets.QApplication(sys.argv)
 # ex = HelpBrowser()
 QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(HelpBrowser.indexFile))

 sys.exit(app.exec())



'''Game Editor
================================
The *game* editor is based on Qt's QTreeWidget. 

|GameEditor|
It has 2 types of lines:

 * regular moves where all columns have a white background
 * beginning of a variant where the middle columns have a grey background

The tree widget has 4 colums:

 #. *Move* shows the actual move or the beginning of a variant in SAN notation 
 #. *Ann* shows the annotation, i.e. a symbolic move assessment
 #. *Pos* shows the position assessment
 #. *Score* shows the engine or material score of the last move in pawns, material score ending with M
 #. *Comment* shows either the move comment or starting comment of a variant
 
By clicking the annotation (*Ann*) and position assessment (*Pos*) fields, a popup dialog
opens which allows to change the contents   

.. |GameEditor| image:: gameEditor.png
  :width: 800
  :alt: Game Editor
'''
from typing import Optional, Set

import sys,  os.path
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets, QtGui, QtCore
else:
 from PyQt6 import QtWidgets, QtGui,  QtCore

import chess, chess.pgn
from chessengine import PGNEval_REGEX
from specialDialogs import ButtonLine, TextEdit, treeWidgetItemPos

class GameTreeView(QtWidgets.QTreeWidget):
 '''Game Editor object
 '''
 itemFlags = QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
 inactiveBrush = QtGui.QBrush(QtGui.QColor('lightgray'))
 colorChars = ('black', 'white')
 endGameSymbols = ['1-0', '0-1', '1/2-1/2', '*']
 endGameDescription = { '1-0'      : 'White wins',
                                  '0-1'       : 'Black wins', 
                                  '1/2-1/2' : 'Drawn game', 
                                  '*'          : 'game in progress' }
 annotationSymbols = ['', '!', '?', '!!', '??', '!?', '?!']
 positionSymbols =  { ''       :  0,
                              '='     : 10, 
                              '~'     : 13, 
                              '+='   : 14, 
                              '=+'   : 15,
                              '+/-' : 16, 
                              '-/+' : 17, 
                              '+-'   : 18,
                              '-+'   : 19,
                              '+--' : 20,
                              '-++' : 21 }
 symbolDescription =  { ''    : 'null annotation',
                              '!'     : 'good move', 
                              '?'     : 'poor move', 
                              '!!'   : 'brilliant move',
                              '??'   : 'blunder',
                              '!?'   : 'intesting move',
                              '?!'   : 'dubious move', 
                              '='     : 'even position', 
                              '~'     : 'unclear position', 
                              '+='   : 'slight advantage for white', 
                              '=+'   : 'slight advantage for black',
                              '+/-' : 'moderate advantage for white', 
                              '-/+' : 'moderate advantage for black', 
                              '+-'   : 'decisive advantage for white',
                              '-+'   : 'decisive advantage for white',
                              '+--' : 'white should resign',
                              '-++' : 'black should resign' }
 piecePawnScoreDict = {
  chess.PAWN : 1, 
  chess.KNIGHT : 3, 
  chess.BISHOP : 3, 
  chess.ROOK : 5, 
  chess.QUEEN : 9
 }

 def __init__(self, parent = None) -> None:
  super(GameTreeView, self).__init__(parent)
  self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
  self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
  self.notifyGameNodeSelectedSignal = None
  self.notifyGameChangedSignal = None
  self._clear()
  self.setColumnCount(5)
  self.setHeaderLabels(['Move', 'Ann', 'Pos', 'Score', 'Comment'])
  headerItem = self.headerItem()
  headerItem.setToolTip(0, 'SAN notation')
  headerItem.setToolTip(1, 'Move assessments')
  headerItem.setToolTip(2, 'Positional assessments')
  headerItem.setToolTip(3, 'Score [pawns] (<float>M == material score)')
  headerItem.setToolTip(4, 'Comment of move or variation')
  self._resetColumnWidth()
  self.clicked.connect(self.on_clicked)
  self.itemExpanded.connect(self.on_itemExpanded)
  self.annotationLine = ButtonLine(self.annotationSymbols, hintDict = self.symbolDescription, pointSize = 12, title = 'Move Annotation', parent = self)
  self.positionLine = ButtonLine(self.positionSymbols, hintDict = self.symbolDescription, pointSize = 12, title = 'Position', parent = self)
  self.endGameLine = ButtonLine(self.endGameSymbols, hintDict = self.endGameDescription, title = 'End Game',  pointSize = 12, parent = self)
  self.commentEdit = TextEdit('Comment ...', pointSize = 10)
  
 def _resetColumnWidth(self) -> None:
  test = QtWidgets.QLabel()
  self.zeroWidth = test.fontMetrics().size(QtCore.Qt.TextFlag.TextSingleLine, '0').width()
  # self.setColumnWidth(1, 20*zeroWidth)
  self.depth = 0
  self.setColumnWidth(0, 15*self.zeroWidth)
  self.setColumnWidth(1, 4*self.zeroWidth)
  self.setColumnWidth(2, 8*self.zeroWidth)
  self.setColumnWidth(3, 8*self.zeroWidth)
  
 @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem)
 def on_itemExpanded(self, widgetItem):
  depth = -1
  while widgetItem is not None:
   depth +=1
   widgetItem = widgetItem.parent()
  self.setColumnWidth(0, (depth*4 + 15)*self.zeroWidth)

 def _clear(self) -> None:
  self.game = None
  self.gameNodeList = list()
  self.gameItemList = list()
  self.gameVariantNodeList = list()
  self.gameVariantItemList = list()
  super(GameTreeView, self).clear()

 def setup(self, notifyGameNodeSelectedSignal : Optional[QtCore.pyqtSignal], 
                       notifyGameChangedSignal : Optional[QtCore.pyqtSignal]) -> None:
  '''Set up of the game editor
  
:param notifyGameNodeSelectedSignal: signal to be emitted if a game node is selected
:param notifyGameChangedSignal: signal to be emitted if the loaded game is changed
  '''
  self.notifyGameNodeSelectedSignal = notifyGameNodeSelectedSignal
  self.notifyGameChangedSignal = notifyGameChangedSignal

 def addVariant(self, gameNode :  chess.pgn.GameNode) -> None:
  '''Adds a new variant, parent node must exist in the editor
  
:param gameNode: game node to be added (must not be main_variation !!)
  '''
  if gameNode is None or gameNode in self.gameVariantNodeList:
   return
  if gameNode.parent not in self.gameNodeList:
   raise ValueError('Parent Node {} not found'.format(gameNode.parent))
  if gameNode.is_main_variation():
   raise ValueError('Node {} is not a variant'.format(gameNode))

  parent = gameNode.parent.variations[0]
  parentIndex = self.gameNodeList.index(parent)
  parentItem = self.gameItemList[parentIndex]
  newVariant = QtWidgets.QTreeWidgetItem()
  for column in range(1, 4):
   newVariant.setBackground(column, self.inactiveBrush)
  newVariant.setFlags(self.itemFlags)
  board = gameNode.parent.board()
  san = board.san(gameNode.move)
  if board.turn:
   moveText = '{}.{}'.format(board.fullmove_number, san)
  else:
   moveText = '... {}'.format(san)
  newVariant.setText(0, '{} ...'.format(moveText))
  newVariant.setText(4, gameNode.starting_comment)
  parentItem.addChild(newVariant)

  self.gameVariantNodeList.append(gameNode)
  self.gameVariantItemList.append(newVariant)

  self.addGameNodes(gameNode, parentItem = newVariant)

 def addGameNodes(self, gameNode :  chess.pgn.GameNode,  
                                     parentItem : Optional[QtWidgets.QTreeWidgetItem] = None) -> None:
  '''Adds 1 or more nodes, parent node of first node must exist in the editor
  
:param gameNode: game node to be added (must be main_variation !!)
:param parentItem: parent item of the gameNode (used only for internal use)
  '''
  if gameNode is None:
   return
  if isinstance(gameNode, chess.pgn.Game):
   self.setGame(gameNode)
   return
  if gameNode.parent not in self.gameNodeList:
   raise ValueError('Parent Node {} not found'.format(gameNode.parent))
  parent = gameNode.parent
  if parentItem is None:
   if gameNode in self.gameNodeList:
    return
   if not gameNode.is_main_variation():
    raise ValueError('Node {} is not in the main variation'.format(gameNode))
   parentIndex = self.gameNodeList.index(parent)
   parentItem = self.gameItemList[parentIndex]
   if parentItem != self.gameItemList[0]:
    parentItem = parentItem.parent()
  board = gameNode.parent.board()
  while gameNode is not None: 
   newNode = QtWidgets.QTreeWidgetItem()
   newNode.setFlags(self.itemFlags)
   parent = gameNode.parent
   san = board.san(gameNode.move)
   if board.turn:
    moveText = '{}.{}'.format(board.fullmove_number, san)
   else:
    moveText = '... {}'.format(san)
   newNode.setText(0, moveText)
   newNode.setText(1, self._findNAGSymbol (True,  gameNode))
   newNode.setText(2, self._findNAGSymbol (False,  gameNode))
   comment = gameNode.comment
   comment = comment.replace('\n','\\n')
   if board.is_checkmate():
    scoreText = 'MATE'
   elif board.is_stalemate():
    scoreText = 'SMATE'
   else:
    scoreText = None
    while True:
     match = PGNEval_REGEX.search(comment)
     if match is None:
      break
     if match.group(1) == '%' or match.group(1) == '%eval':
      scoreText = match.group(2)
     comment = comment.replace(match.group(0),'')
    if scoreText is None:
     pieceMap = gameNode.board().piece_map()
     pawnScore = 0
     for piece in list(pieceMap.values()):
      if piece.piece_type != chess.KING:
       if piece.color == chess.WHITE:
        pawnScore += self.piecePawnScoreDict[piece.piece_type]
       else:
        pawnScore -= self.piecePawnScoreDict[piece.piece_type]
     scoreText = '{}M'.format(pawnScore)
   newNode.setText(3, scoreText)
   newNode.setText(4, comment)
   
   parentItem.addChild(newNode)
   self.gameNodeList.append(gameNode)
   self.gameItemList.append(newNode)
   for variantNode in parent.variations[1:]:
    self.addVariant(variantNode)
    
   board.push(gameNode.move)
   gameNode = gameNode.next()
   
 def removeGameNode(self, gameNode : chess.pgn.GameNode) -> None: 
  if not gameNode.is_end() or gameNode.parent is None or gameNode not in self.gameNodeList:
   return
  index = self.gameNodeList.index(gameNode)
  self.gameItemList[index].parent().removeChild(self.gameItemList[index])
  self.gameNodeList.pop(index)
  self.gameItemList.pop(index)
  
 def setGameResult(self, result : str) -> None:
  '''Sets the 'Result' header element

:param result: one out of *1-0*, *0-1*, *1/2-1/2*, *\**
  '''
  gameResult = ["1-0", "0-1", "1/2-1/2", "*"]
  if result not in gameResult:
   result = gameResult[3]
  if len(self.gameItemList) > 0:
   self.gameItemList[0].setText(0, 'Result: {}'.format(result))

   
 def setGame(self, game : chess.pgn.Game) -> None:
  '''Clears the editor and sets new game
  
:param game: game to be set
  '''
  self._clear()
  self.game = game
  master = QtWidgets.QTreeWidgetItem()
  master.setFlags(self.itemFlags)
  master.setText(4, self.game.comment)
  self.addTopLevelItem(master)
  self.gameNodeList = [self.game]
  self.gameItemList = [master]
  self.setGameResult(self.game.headers['Result'])
  self.addGameNodes(self.game.next(), master)
  master.setExpanded(True)

 def selectNodeItem(self, gameNode : chess.pgn.GameNode) -> None:
  '''Selects a game node
  
:param gameNode: game node to be selected
  '''
  if gameNode not in self.gameNodeList:
   selIndex = 0
  else:
   selIndex = self.gameNodeList.index(gameNode)
  if selIndex > 0:
   self.expandItem(self.gameItemList[selIndex])
  self.setCurrentItem(self.gameItemList[selIndex], 0, QtCore.QItemSelectionModel.SelectionFlag.SelectCurrent)

 def selectSubnodeItem(self, gameNode : chess.pgn.GameNode, next : bool = True):
  '''Selects the next or previous variant
  
:param gameNode: reference game node
:param next: if True select the next variant, else select previous variant
  '''
  if gameNode not in self.gameNodeList:
   raise ValueError('selectSubnodeItem/gameNode: {} not found'.format(gameNode))
  parentNode = gameNode.parent
  if parentNode is None:
   return
  if not next:
   # go back to the first variation
   newNodeIndex = parentNode.variations.index(gameNode) - 1
   if newNodeIndex >= 0:
    self.selectNodeItem(parentNode.variations[newNodeIndex])
  else:
   # go to next variation
   newNodeIndex = parentNode.variations.index(gameNode) + 1
   if newNodeIndex < len(parentNode.variations):
    selIndex = self.gameNodeList.index(gameNode)
    self.expandItem(self.gameItemList[selIndex])
    newGameNode = parentNode.variations[newNodeIndex]
    self.selectNodeItem(newGameNode)
    self.notifyGameNodeSelectedSignal.emit(newGameNode)
    
 def _findNAGSymbol(self, isAnnotation : bool,  gameNode : chess.pgn.GameNode) -> str:
  if isAnnotation:
   for nag, sym in enumerate(self.annotationSymbols):
    if nag in gameNode.nags:
     return sym
  else:
   for sym, nag in self.positionSymbols.items():
    if nag in gameNode.nags:
     return sym

 def _updateNAGs(self, isAnnotation : bool,  gameNode : chess.pgn.GameNode,  nag : int) -> Set[int]:
  newNAGs = set()
  if nag != 0:
   if isAnnotation:
    r = range(1, 10)
   else:
    r = range(10, 140)
   for _nag in gameNode.nags:
    if _nag not in r:
     newNAGs .add(_nag)
   newNAGs .add(nag)
  return newNAGs
  
 def _editComment(self, item :  QtWidgets.QTreeWidgetItem) -> str:
  self.commentEdit.setText(item.text(4).replace('\\n','\n'))
  if not self.commentEdit.exec():
   return None
  comment = self.commentEdit.text().replace('\n','\\n')
  item.setText(4, comment)
  score = item.text(3)
  if len(score) > 0 and score[-1] != 'M':
   comment = '[%eval {}] {}'.format(score, comment)
  return comment

 @QtCore.pyqtSlot(QtCore.QModelIndex)
 def on_clicked(self, index):
  item = self.itemFromIndex(index)
  column = index.column()
  isVariant = False
  parent = index
  while parent.isValid():
   isVariant = not isVariant
   parent = parent.parent()
  if isVariant:
   if column == 4:
    if item in self.gameVariantItemList:
     selIndex = self.gameVariantItemList.index(item)
     gameNode = self.gameVariantNodeList[selIndex]
     gameNode.starting_comment = self._editComment(item)
    else:
     self.game.comment = self._editComment(item)
     gameNode = self.game
   else:
    if column == 0 and self.notifyGameNodeSelectedSignal is not None:
     self.notifyGameNodeSelectedSignal.emit(self.game)
    return
  else:
   if item in self.gameItemList:
    selIndex = self.gameItemList.index(item)
    gameNode = self.gameNodeList[selIndex]
   else:
    return
   if column == 0:
    if gameNode is None:
     self.endGameLine.setFocus('*')
     self.endGameLine.move(treeWidgetItemPos(item))
     endGameID = self.endGameLine.exec()
     self.game.headers['Result'] = self.endGameSymbols[endGameID]
     item.setText(column, self.endGameSymbols[endGameID])
    else:
     if self.notifyGameNodeSelectedSignal is not None:
      self.notifyGameNodeSelectedSignal.emit(gameNode)
     return
   elif column == 1:
    self.annotationLine.move(treeWidgetItemPos(item))
    nag = self.annotationLine.exec()
    gameNode.nags = self._updateNAGs(True, gameNode, nag)
    item.setText(column, self._findNAGSymbol (True,  gameNode))
   elif column == 2:
    self.positionLine.move(treeWidgetItemPos(item))
    nag = self.positionLine.exec()
    gameNode.nags = self._updateNAGs(True, gameNode, nag)
    item.setText(column, self._findNAGSymbol (False,  gameNode))
   elif column == 4:
    gameNode.comment = self._editComment(item)
   else:
    return
  self._emitGameChanged()
  
 def _isVariant(self, item : QtWidgets.QTreeWidgetItem) -> bool:
  isTrue = False
  parent = item
  while parent is not None:
   isTrue = not isTrue
   parent = parent.parent()
  return isTrue

 def removeVariant(self) -> None:
  '''Remove variant, if currently selected item is a variant
  '''
  item = self.currentItem()
  if item is not None and self._isVariant(item):
   parent = item.parent()
   if parent is None:
    return
   index = parent.indexOfChild(item)
   parent.removeChild(item)
   self.setCurrentItem(parent.child(index))
   selIndex = self.gameVariantItemList.index(item)
   itemNode = self.gameVariantNodeList[selIndex]
   itemNode.parent.remove_variation(itemNode)
   self._emitGameChanged()

 def promoteVariant(self) -> None:
  '''Promote variant, if currently selected item is a variant
  '''
  item = self.currentItem()
  if item is not None and self._isVariant(item):
   parent = item.parent()
   if parent is None:
    return
   index = parent.indexOfChild(item)
   if index > 0:
    parent.takeChild(index)
    parent.insertChild(index - 1, item)
    self.setCurrentItem(item)
    parent.setExpanded(True)
    item.setExpanded(False)
    selIndex = self.gameVariantItemList.index(item)
    itemNode = self.gameVariantNodeList[selIndex]
    itemNode.parent.promote(itemNode)
    self._emitGameChanged()

 def demoteVariant(self) -> None:
  '''Demote variant, if currently selected item is a variant
  '''
  item = self.currentItem()
  if item is not None and self._isVariant(item):
   parent = item.parent()
   if parent is None:
    return
   index = parent.indexOfChild(item)
   if index < parent.childCount():
    parent.takeChild(index)
    parent.insertChild(index + 1, item)
    self.setCurrentItem(item)
    parent.setExpanded(True)
    item.setExpanded(False)
    selIndex = self.gameVariantItemList.index(item)
    itemNode = self.gameVariantNodeList[selIndex]
    itemNode.parent.demote(itemNode)
    self._emitGameChanged()

 def promoteVariant2Main(self) -> None:
  '''Promote variant to main line, if currently selected item is a variant
  '''
  item = self.currentItem()
  if item is not None:
   parent = item.parent()
   if self._isVariant(item):
    index = parent.indexOfChild(item)
    if index > 0:
     child = parent.takeChild(index)
     parent.insertChild(0, child)
    vChildren = item.takeChildren()
    gParent = parent.parent()
    fIndex = gParent.indexOfChild(parent)
    pChildren = parent.takeChildren()
    mChildren = list()
    for idx in reversed(range(fIndex, gParent.childCount())):
     mChildren.insert(0, gParent.takeChild(idx))
    gParent.addChildren(vChildren)
    vChildren[0].addChildren(pChildren)
    item.addChildren(mChildren)
    self.setCurrentItem(vChildren[0])
    parent.setExpanded(True)
    item.setExpanded(False)
    selIndex = self.gameVariantItemList.index(item)
    itemNode = self.gameVariantNodeList[selIndex]
    itemNode.parent.promote_to_main(itemNode)
    self._emitGameChanged()
  
 def _emitGameChanged(self) -> None:
  if self.notifyGameChangedSignal is not None:
   self.notifyGameChangedSignal.emit(self.game)

if __name__ == "__main__":
 import io, sys
 from pgnParse import read_game

 if False:
  newData = """[Event "matein2"]
[Site "problem solved"]
[Date "????.??.??"]
[Round "?"]
[White "?"]
[Black "?"]
[Result "*"]
[Comment "1"]
[PlyCount "0"]
[FEN "1k6/Rp1K4/1P5P/8/P7/3pP3/1p1P4/8 w - - 0 1"]
  
1.h7 b1=Q! $20 {[% -4.80]} 2.h8=R# *"""
 else:
  import os.path
  fileDirectory = os.path.dirname(os.path.abspath(__file__))
  ps = os.path.join(fileDirectory, 'training', 'openings', 'kingsPawn.pgn')
  with open(ps, mode = 'r',  encoding = 'utf-8') as f:
   newData = f.read()
 
 pgn = io.StringIO(newData)
 game = read_game(pgn)
 app = QtWidgets.QApplication([])

 tree = GameTreeView()
 
 msgSc = QtGui.QShortcut(QtGui.QKeySequence('Ctrl+U'), tree)
 msgSc.activated.connect(tree.promoteVariant)

 msgSc = QtGui.QShortcut(QtGui.QKeySequence('Ctrl+D'), tree)
 msgSc.activated.connect(tree.demoteVariant)

 msgSc = QtGui.QShortcut(QtGui.QKeySequence('Ctrl+X'), tree)
 msgSc.activated.connect(tree.removeVariant)

 msgSc = QtGui.QShortcut(QtGui.QKeySequence('Ctrl+M'), tree)
 msgSc.activated.connect(tree.promoteVariant2Main)

 tree.setGame(game)
 tree.resize(500,400)
 tree.show()
 sys.exit(app.exec())



'''
Configuration of Engines
================================

This dialog allows to add/remove/configure engines

|ConfigureEngine| 

Configuration of the Engine's Options
===========================================

By pressing the *Details* button the *Configure Engine Options* dialog is started.
Please consult the engine's documentation (e.g. `stockfish`_ or `here`_) before changing any options.

|ConfigureEngineOptions| 

.. |ConfigureEngine| image:: configureEngine.png
  :width: 400
  :alt: Configure Engine
.. |ConfigureEngineOptions| image:: configureEngineOptions.png
  :width: 400
  :alt: Configure Engine Options
.. _stockfish: https://github.com/official-stockfish/Stockfish
.. _here: https://www.chessprogramming.org/Engines
'''
from typing import List, Dict, Callable, Union, Optional, Any
import sys, os, os.path
import platform
import copy

import configparser
import re

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets, QtCore
 from PyQt5 import uic
else:
 from PyQt6 import QtWidgets, QtCore
 from PyQt6 import uic

from chessengine import ChessEngine
from configureEngineOptions import ConfigureEngineOptions 

intRe = re.compile(r"^[+\-]?[0-9]+$")
boolRe = re.compile(r"^(True|False)$")

def loadEngineSettings(settings : configparser.ConfigParser) -> Dict[str, List[Any]]:
 engineDict = dict()
 for section in settings.sections():
  partList = section.split('/')
  isEngineSection = (len(partList) == 2 and partList[0] == 'Engine')
  if  isEngineSection:
   key = partList[1]
   optionsDict = dict()
   executable = None
   for opt in settings[section]:
    value = settings[section][opt]
    if opt == 'executable':
     executable = value
    elif intRe.match(value) is not None:
     optionsDict[opt] = int(value)
    elif boolRe.match(value) is not None:
     optionsDict[opt] = bool(value)
    else:
     optionsDict[opt] = value
   if executable is not None:
    if os.path.exists(executable):
     engineDict[key] = [executable, optionsDict]
    else:
     print('loadEngineSettings: {} not found'.format(executable))
 return engineDict

def saveEngineSettings(settings : configparser.ConfigParser, engineDict : Dict[str, List[Any]]) -> None:
 for section in settings.sections():
  partList = section.split('/')
  if len(partList) == 2 and partList[0] == 'Engine':
   settings.remove_section(section)
 for opt in engineDict:
  executable, optionsDict = engineDict[opt]
  key = 'Engine/{}'.format(opt)
  settings[key] = optionsDict
  settings[key]['executable'] = executable

class ConfigureEngine(QtWidgets.QDialog):
 fileDialogOptions = QtWidgets.QFileDialog.Option.DontUseNativeDialog
 fileDialogFilters = QtCore.QDir.Filter.AllDirs \
                      | QtCore.QDir.Filter.Files \
                      | QtCore.QDir.Filter.NoDotAndDotDot \
                      | QtCore.QDir.Filter.Hidden

 def __init__(self, parent : Optional[QtCore.QObject] = None):
  super(ConfigureEngine, self).__init__(parent)
  fileDirectory = os.path.dirname(os.path.abspath(__file__))
  uic.loadUi(os.path.join(fileDirectory, 'configureEngine.ui'), self)

  self.engineDict = dict()
  self.log = False
  self.directories = list()
  self.engineTableWidget.horizontalHeader().setStretchLastSection(True)
  self.engineTableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
  self.browsePushButton.setIcon(self.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_DialogSaveButton))
  self.engineDetails = dict()

 @QtCore.pyqtSlot(str)
 def notifyError(self, msg):
  msgBox = QtWidgets.QMessageBox()
  msgBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
  msgBox.setText(msg)
  msgBox.setWindowTitle("Error")
  msgBox.exec()

 def _fillForm(self):
  self.engineTableWidget.setRowCount(len(self.engineDict))
  for row, n2ec in enumerate(self.engineDict.items()):
   name, ec = n2ec
   executable = ec[0]
   if not (os.path.isfile(executable) and os.access(executable, os.X_OK )):
    executable = 'not found'
   else:
    self.directories.append(os.path.dirname(executable))
   self.engineTableWidget.setCellWidget(row, 0, QtWidgets.QLabel(name))
   self.engineTableWidget.setCellWidget(row, 1, QtWidgets.QLabel(executable))
  self.executableLineEdit.setText('')
  self.nameLineEdit.setText('')

 def run(self, engineDict : Dict[str, List[Any]] = dict(),  log : Union[Callable, bool] = False) -> Dict[str, List[Any]]:
  self.engineDict = copy.deepcopy(engineDict)
  self.log = log
  self._fillForm()
  self.isChanged = False
  self.selectedRow = None
  if self.exec():
   return self.engineDict
  else:
   return engineDict
  
 def ignoreIsChanged(self) -> None:
  if self.isChanged:
   msgBox = QtWidgets.QMessageBox()
   msgBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
   msgBox.addButton(QtWidgets.QMessageBox.StandardButton.No)
   msgBox.addButton(QtWidgets.QMessageBox.StandardButton.Yes)
   msgBox.setText('Ignore ?')
   msgBox.setWindowTitle("Unsaved content detected")
   rc = msgBox.exec()
   return rc
  else:
   return True

 def checkEngine(self, executable : str) -> Optional[ChessEngine]:
  try:
   engine = ChessEngine([executable, dict()], log = self.log)
   isOK = engine.isReady()
  except:
   isOK = False
  if not isOK:
   msgBox = QtWidgets.QMessageBox()
   msgBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
   msgBox.addButton(QtWidgets.QMessageBox.StandardButton.Ok)
   msgBox.setText('{} is not an UCI engine'.format(executable))
   msgBox.setWindowTitle("Error")
   msgBox.exec()
   return None
  return engine
 
 @QtCore.pyqtSlot()
 def on_detailsPushButton_clicked(self):
  executable = self.executableLineEdit.text()   
  if len(executable) == 0:
   self.notifyError('Executable undefined')
   return
  name = self.nameLineEdit.text()
  engine = self.checkEngine(executable)
  if engine is not None:
   dialog = ConfigureEngineOptions()
   newEngineDetails = dialog.run(engine.optionsDict, changedOptions = self.engineDetails, name = name)
   if newEngineDetails != self.engineDetails:
    self.engineDetails = newEngineDetails
    self.isChanged = True

 @QtCore.pyqtSlot()
 def on_browsePushButton_clicked(self):
  fDialog = QtWidgets.QFileDialog()
  fDialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFile)
  fDialog.setOptions(self.fileDialogOptions)
  fDialog.setFilter(self.fileDialogFilters)
  fDialog.setWindowTitle("Load Engine Executable ...")
  if len(self.directories) > 0:
   fDialog.setHistory(self.directories)
  if len(self.executableLineEdit.text()) > 0:
   head, tail = os.path.split(self.executableLineEdit.text())
   fDialog.setDirectory(head)
   fDialog.selectFile(tail)
  if platform.system() == 'Windows':
   fDialog.setNameFilter("Executables (*.exe);;All Files (*)")
  else:
   fDialog.setNameFilter("All Files (*)")
  if fDialog.exec():
   executable = fDialog.selectedFiles()[0]
   self.engineDetails = dict()
   if self.checkEngine(executable) is not None:
    self.executableLineEdit.setText(executable)
    self.isChanged = True

 @QtCore.pyqtSlot(int, int)
 def on_engineTableWidget_cellClicked(self, row, column):
  if not self.ignoreIsChanged():
   return  
  name = self.engineTableWidget.cellWidget(row, 0).text()
  self.nameLineEdit.setText(name)
  self.executableLineEdit.setText(self.engineDict[name][0])
  self.engineDetails = self.engineDict[name][1]
  self.selectedRow = row

 @QtCore.pyqtSlot(str)
 def on_nameLineEdit_textEdited(self, newName):
  self.isChanged = True

 @QtCore.pyqtSlot(str)
 def on_executableLineEdit_textEdited(self, newExecutable):
  if self.checkEngine(newExecutable) is None:
   self.executableLineEdit.setText('')
   self.engineDetails = dict()
  self.isChanged = True

 @QtCore.pyqtSlot()
 def on_savePushButton_clicked(self):
  name = self.nameLineEdit.text()
  executable = self.executableLineEdit.text()
  if not self.isChanged:
   return
  if len(executable) == 0:
   self.notifyError('Executable undefined')
   return
  self.engineDict[name] = [executable, self.engineDetails]
  self._fillForm()
  self.isChanged = False
 
 @QtCore.pyqtSlot()
 def on_removePushButton_clicked(self):
  if not self.ignoreIsChanged():
   return
  key = self.engineTableWidget.cellWidget(self.selectedRow, 0).text()
  self.engineDict.pop(key, None)
  self.engineTableWidget.removeRow(self.selectedRow)
  self.selectedRow = None
  
 @QtCore.pyqtSlot()
 def on_cancelPushButton_clicked(self):
  self.done(False)
  
 @QtCore.pyqtSlot()
 def on_okPushButton_clicked(self):
  if self.ignoreIsChanged():
   self.done(True)

if __name__ == "__main__":
 import os,  os.path
 import sys
 import configparser

 qApp=QtWidgets.QApplication(sys.argv)   

 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 os.chdir(fileDirectory)
 
 settingsFile = os.path.join(fileDirectory, 'settings.ini')
 settings = configparser.ConfigParser(delimiters=['='], allow_no_value=True)
 settings.optionxform = str
 settings.read(settingsFile, encoding = 'utf-8')
 engineDict = loadEngineSettings(settings)

 configureEngine = ConfigureEngine()
 print(engineDict)
 for name in engineDict:
  engineDict[name] = [engineDict[name], dict()] 
 configureEngine.show()
 newEngineDict = configureEngine.run(engineDict = engineDict, log = True)
 print('------- new')
 print(newEngineDict)
 qApp.exec()



__description__ = \
"""
A PyQt6 based chess GUI

Versions:
 1.0.0      first version
 1.0.3      installation running
 1.0.4      bug fixes + end-of-game handling completed
 1.0.6      documentation on github.io + ico added
 1.0.8      setup corrected
 1.1.0      FEN-builder added and bug fixes
 1.2.0      tests established and bug fixes
 1.2.1      variant bug fixes
 1.2.2      bug fixes
 1.2.3      bug fixes
 1.3.0      migrated to uic.loadUI
 1.4.0      migrated to PyQt6, Open/Save issues removed
 1.5.0      migrated to combined PyQt6/PyQt5 operation
 2.0.0      migrated to setup.cfg 
              postInstall.py added
              flexible gamelisttableview implemented
              detailed hint mechanism added
"""
__author__ = "Reinhard Maerz"
__date__ = "2022-08-18"
__version__ = "2.0.0"

__all__ = [
 'AboutDialog', 
 'AnnotateEngine', 'Annotator', 
 'ChessMainWindow', 'runMzChess', 
 'BuildFenClass', 'runFenBuilder', 
 'installLeipFont'
 'ChessEngine', 
 'ConfigureEngine', 'loadEngineSettings', 'saveEngineSettings', 
 'ConfigureEngineOptions',  
 'ECODatabase', 'TSVType', 
 'GameHeaderView', 'KeyType', 
 'GameListTableModel', 'GameListTableView', 
 'HelpBrowser', 
 'checkFEN','read_game', 'read_board', 'read_headers', 'skip_game', 'PGNLexer', 
 'QBoardViewClass', 'Piece', 'Game', 
 'ScorePlot', 
 'ButtonLine', 'ItemSelector', 'treeWidgetItemPos', 
 'QUCIEdit', 'UCIHighlighter', 
 'warnOfDanger'
]


import os.path, sys
def useQt5():
 try:
  import PyQt6.QtSvgWidgets
  import PyQt6.QtCharts
  return False
 except:
  import PyQt5.QtSvg
  import PyQt5.QtChart
  return True

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from postInstall import postInstall
from .AboutDialog import AboutDialog
from .annotateEngine import AnnotateEngine,  Annotator
from .installLeipFont import installLeipFont
from .chessengine import ChessEngine
from .configureEngine import ConfigureEngine, loadEngineSettings, saveEngineSettings
from .configureEngineOptions import ConfigureEngineOptions
from .eco import ECODatabase, TSVType
from .gameheaderview import GameHeaderView, KeyType
from .gamelisttableview import GameListTableModel, GameListTableView
from .helpDialog import HelpBrowser
from .pgnParse import checkFEN, read_game, read_board, read_headers, skip_game, PGNLexer
from .qboardviewclass import QBoardViewClass, Piece, Game
from .scoreplotgraphicsview import ScorePlot
from .specialDialogs import ButtonLine, ItemSelector, treeWidgetItemPos
from .uciedit import QUCIEdit, UCIHighlighter
from .warnOfDanger import warnOfDanger
from .qbuildfen import BuildFenClass, SelectionBox, PlacementBoard, runFenBuilder
from .chessMainWindow import ChessMainWindow, runMzChess



from typing import Dict, Tuple, Any
import pytest

inverseLeipzigEncodeDict = dict()

from PyQt6 import QtCore

import chess, chess.pgn
import MzChess

@pytest.fixture()
def app(qtbot):
 global inverseLeipzigEncodeDict
 inverseLeipzigEncodeDict = dict()
 for key, lKey in MzChess.Game.leipzigEncodeDict.items():
  inverseLeipzigEncodeDict[lKey] = key
 mzChess = MzChess.ChessMainWindow()
 mzChess.show()
 mzChess.setup()
 qtbot.addWidget(mzChess)
 return mzChess

def materialBudget(board : chess.Board) -> str:
 pieceMap = board.piece_map()
 whiteCountList = 7 * [0]
 blackCountList = 7 * [0]
 for piece in list(pieceMap.values()):
  if piece.color == chess.WHITE:
   whiteCountList[piece.piece_type] += 1
  else:
   blackCountList[piece.piece_type] += 1
 pieceText =''
 for pieceType in range(5, 0,-1):
  delta = whiteCountList[pieceType] - blackCountList[pieceType]
  if delta == 0:
   continue
  if delta > 0:
   pieceText += delta * chess.Piece(pieceType, chess.WHITE).symbol()
  else:
   pieceText += abs(delta) * chess.encodePiece(pieceType, chess.BLACK).symbol()
 return pieceText
  
def makeMove(app, qtbot, 
                      move : chess.Move, 
                      breakAfterPress : bool = False, 
                      breakAfterRelease : bool = False, 
                      relDeltaPercent : Tuple[bool, QtCore.QPoint] = (False, QtCore.QPoint()),  
                      expectedResultDict : Dict[str, Any] = dict()) -> None:
 global inverseLeipzigEncodeDict
 gv = app.boardGraphicsView.viewport()
 gameScene = app.boardGraphicsView.game
 if relDeltaPercent[0]:
  delta = (0.01 * gameScene.pieceSize) * relDeltaPercent[1]
 else:
  delta = QtCore.QPoint()
 fromPoint = gameScene.getScenePos(move.from_square).toPoint() + delta
 qtbot.mousePress(gv, QtCore.Qt.MouseButton.LeftButton, pos = app.boardGraphicsView.mapFromScene(fromPoint))
 if breakAfterPress:
  breakpoint()
 if relDeltaPercent[0]:
  delta = QtCore.QPoint()
 else:
  delta = (0.01 * gameScene.pieceSize) * relDeltaPercent[1]
 toPoint = gameScene.getScenePos(move.to_square).toPoint() + delta
 qtbot.mouseRelease(gv, QtCore.Qt.MouseButton.LeftButton, pos = app.boardGraphicsView.mapFromScene(toPoint))
 QtCore.QCoreApplication.processEvents()
 if breakAfterRelease:
  breakpoint()
 if 'FEN' in expectedResultDict:
  assert expectedResultDict['FEN'] == app.gameNode.board().fen()
 if 'Material' in expectedResultDict:
  pieceText =''
  for char in gameScene.materialLabel.text():
   pieceText += inverseLeipzigEncodeDict[char]
  for char in 'QRBKPqrbkp':
   assert expectedResultDict['Material'].count(char) == pieceText.count(char)
 if 'Color' in expectedResultDict:
  board = app.gameNode.board()
  if board.is_game_over():
   next = None
  else:
   next = board.turn
  assert expectedResultDict['Color'] == next
 
def runPromoteMate(app, qtbot, promoteMateGame : chess.pgn.Game) -> None:
 app.actionNew_Game.triggered.emit()
 expectedResultDict = dict()
 gameNode = promoteMateGame.next()
 print('You must promote a queen for this example.')
 while gameNode is not None:
  board = gameNode.board()
  expectedResultDict['Material'] = materialBudget(board)
  expectedResultDict['FEN'] = board.fen()
  if gameNode.board().is_game_over():
   expectedResultDict['Color'] = None
  else:
   expectedResultDict['Color'] = gameNode.turn()
  makeMove(app, qtbot, gameNode.move, 
                        breakAfterPress = app.actionShow_Options.isChecked(), 
                        breakAfterRelease = app.actionWarn_of_Danger.isChecked(), 
                        expectedResultDict = expectedResultDict)
  gameNode = gameNode.next()

def test_promoteMate1(app, qtbot, promoteMateGame : chess.pgn.Game) -> None:
 app.actionFlip_Board.setChecked(False)
 app.actionShow_Options.setChecked(True)
 app.actionWarn_of_Danger.setChecked(False)
 runPromoteMate(app, qtbot, promoteMateGame)
 
def test_promoteMate2(app, qtbot, promoteMateGame : chess.pgn.Game) -> None:
 app.actionFlip_Board.setChecked(True)
 app.actionShow_Options.setChecked(False)
 app.actionWarn_of_Danger.setChecked(True)
 runPromoteMate(app, qtbot, promoteMateGame)
 
def mouseOp(app, qtbot, analyzePress : bool) -> None:
 app.actionNew_Game.triggered.emit()
 mouseOpStr = ('mouseReleaseEvent', 'mousePressEvent')
 nPoints = 3
 expectedResultDict = {'FEN' : 'rnbqkbnr/pppppppp/8/8/4P3/8/PPPP1PPP/RNBQKBNR b KQkq - 0 1'}
 gameScene = app.boardGraphicsView.game
 squarePos = gameScene.getScenePos(chess.parse_square('e2'))
 print('{}: pos(e2) = {}'.format(mouseOpStr[analyzePress], squarePos))
 expectedResultDict = {}
 for nx in range(nPoints):
  for ny in range(nPoints):
   relDeltaPercent = QtCore.QPoint((100 * nx) // nPoints, (100 * ny) // nPoints)
   delta = (0.01 * gameScene.pieceSize) * relDeltaPercent
   print('{}: actPos = {}'.format((nx, ny), squarePos + delta))
   makeMove(app, qtbot, chess.Move.from_uci('e2e4'), 
                   breakAfterPress = False, 
                   breakAfterRelease = False, 
                   relDeltaPercent = (analyzePress, relDeltaPercent), 
                   expectedResultDict = expectedResultDict)
   app.actionUndo_Last_Move.triggered.emit()

def test_mousePress(app, qtbot, promoteMateGame : chess.pgn.Game) -> None:
 mouseOp(app, qtbot, False)

def test_mouseRelease(app, qtbot, promoteMateGame : chess.pgn.Game) -> None:
 mouseOp(app, qtbot, False)

def test_exitMzChess():
 pytest.helpers.exitApp()



import sys,  os.path
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets, QtCore
 from PyQt5.QtWidgets import QAction
else:
 from PyQt6 import QtWidgets, QtCore
 from PyQt6.QtGui import QAction

class GameListTableModel(QtCore.QAbstractTableModel):
 
 def __init__(self, gameList, gameHeaderKeys, parent = None):
  super(GameListTableModel, self).__init__(parent)
  self.gameHeaderKeys = gameHeaderKeys
  self.gameList = gameList

 def rowCount(self, parent = None):
  return len(self.gameList)

 def columnCount(self, parent = None):
  return len(self.gameHeaderKeys)

 def data(self, index, role = QtCore.Qt.ItemDataRole.DisplayRole):
  if index.isValid():
   if role == QtCore.Qt.ItemDataRole.DisplayRole:
    actGameHeaders = self.gameList[index.row()].headers
    columnKey = self.gameHeaderKeys[index.column()]
    return actGameHeaders[columnKey]
   if role == QtCore.Qt.ItemDataRole.TextAlignmentRole:
    return QtCore.Qt.AlignmentFlag.AlignCenter | QtCore.Qt.AlignmentFlag.AlignVCenter
  return None

 def headerData(self, col, orientation, role = QtCore.Qt.ItemDataRole.DisplayRole):
  if role == QtCore.Qt.ItemDataRole.DisplayRole:
   if orientation == QtCore.Qt.Orientation.Horizontal:
    return self.gameHeaderKeys[col]
   if orientation == QtCore.Qt.Orientation.Vertical:
    return '#{0:}'.format(col)
  return None

class GameListTableView(QtWidgets.QTableView):
 sevenTagRoster = ["Event", "Site", "Round", "Date", "White", "Black", "Result"]

 def __init__(self, parent = None):
  super(GameListTableView, self).__init__(parent)
  self.doubleClicked.connect(self.on_doubleClicked)
  self.horizontalHeader().setStretchLastSection(True)
  self.horizontalScrollBar().setDisabled(True)
  self.notifyDoubleClickSignal = None
  self.notifyHeaderChangedSignal = None
  self.sizeHints = None
  self.gameHeaderKeys = ["Date", "White", "Black", "Result"]
  headerWidget = self.horizontalHeader()
  headerWidget.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
  headerWidget.customContextMenuRequested.connect(self.on_menuHorizontalHeader_requested)
  self.context = QtWidgets.QMenu(self)
  self.context.triggered.connect(self.on_menuContext_triggered)
  for tag in self.sevenTagRoster:
   self.context.addAction('Show {}'.format(tag))
  self.context.addSeparator()
  self.context.addAction('Hide')
  
 @QtCore.pyqtSlot(QAction)
 def on_menuContext_triggered(self, action):
  actionText = action.text()
  newGameHeaderKeys = self.gameHeaderKeys.copy()
  if actionText == 'Hide':
   if len(newGameHeaderKeys) <= 1:
    self.notifyError('At least 1 column required')
    return
   newGameHeaderKeys.pop(self.contextColumn)
  else:
   item = actionText.split(' ')[1]
   newGameHeaderKeys.insert(self.contextColumn, item)
  self.gameHeaderKeys = newGameHeaderKeys
  if self.notifyHeaderChangedSignal is not None:
   self.notifyHeaderChangedSignal.emit(newGameHeaderKeys)
  
 def on_menuHorizontalHeader_requested(self, pos):
  self.contextColumn = self.columnAt(pos.x())
  self.context.exec(self.mapToGlobal(pos))
  return
  self.popupMenu = QtWidgets.QMenu()
  self.popupMenu.exec()
  return
 
 def notifyError(self, str : str) -> None:
  msgBox = QtWidgets.QMessageBox()
  msgBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
  msgBox.setText(str)
  msgBox.setWindowTitle("Error ...")
  msgBox.exec()

 @property
 def gameHeaderKeys(self):
  return self._gameHeaderKeys

 @gameHeaderKeys.setter
 def gameHeaderKeys(self, newGameHeaderKeys):
  if len(newGameHeaderKeys) == 0:
   self.notifyError('len(newGameHeaderKeys) > 0')
   return
  for tag in newGameHeaderKeys:
   if tag not in self.sevenTagRoster:
    self.notifyError('{} not in {}'.format(tag, self.sevenTagRoster))
    return     
  self._gameHeaderKeys = newGameHeaderKeys
  model = self.model()
  if model is not None:
   model.beginResetModel()
   model.gameHeaderKeys = self._gameHeaderKeys 
   model.endResetModel()
  self.contextColumn = -1
  return 

 def setGameList(self, gameList):
  model = GameListTableModel(gameList, self.gameHeaderKeys)
  self.setModel(model)
  self.sizeHints = 4*[0]
  totSize = 0
  for column, key in enumerate(model.gameHeaderKeys):
   self.sizeHints[column] = self.fontMetrics().size(QtCore.Qt.TextFlag.TextSingleLine, key).width()
   totSize += self.columnWidth(column)
   for n, game in enumerate(gameList):
    if key not in game.headers:
     raise ValueError('GameListTableModel: key {} not in game #{}'.format(key,n))
    actSize = self.fontMetrics().size(QtCore.Qt.TextFlag.TextSingleLine, game.headers[key]).width()
    self.sizeHints[column] = max(self.sizeHints[column], actSize)
  self.setColumnWidths()
  
 def setup(self, notifyDoubleClickSignal = None, notifyHeaderChangedSignal = None):
  self.notifyDoubleClickSignal = notifyDoubleClickSignal
  self.notifyHeaderChangedSignal = notifyHeaderChangedSignal
  
 def resetDB(self):
  self.reset()
 
 def setColumnWidths(self):
  if self.sizeHints is None:
   return
  self.setColumnWidth(0, int(self.sizeHints[0]*1.2))
  #self.setColumnWidth(1, middleWidth)
  #self.setColumnWidth(2, middleWidth)
  self.setColumnWidth(3, self.sizeHints[3])

 def resizeEvent(self, ev):
  QtWidgets.QAbstractItemView.resizeEvent(self, ev)
  self.setColumnWidths()
  
 @QtCore.pyqtSlot(QtCore.QModelIndex)
 def on_doubleClicked(self, index):
  if self.notifyDoubleClickSignal is not None:
   self.notifyDoubleClickSignal.emit(index.row())
  else:
   print('game #{} selected'.format(index.row()))

if __name__ == "__main__":
 from pgnParse import read_game

 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 pgnName = os.path.join(fileDirectory, 'training', 'matein', 'matein1.pgn')
 pgn = open(pgnName, mode = 'r',  encoding = 'utf-8')
 gameList = list()
 for n in range(10):
  game = read_game(pgn)
  if game is None:
   break
  gameList.append(game)
 app = QtWidgets.QApplication([])

 gameView = GameListTableView()
 gameView.show()
 gameView.setGameList(gameList)
 gameView.gameHeaderKeys = ['Event', 'Date']
 
 sys.exit(app.exec())



#!/usr/bin/python3

import sys
import itertools


def main(argv):
    name = "X"
    for prefix in "abcde":
        if f"{prefix}.tsv" in argv[1]:
            name = prefix.upper()

    print("package chess")
    print("package opening")
    print()
    print("// Generated from https://github.com/niklasf/eco")
    print("// format: off")
    print(f"private[opening] object FullOpeningPart{name} {{")
    print()
    print("  def db: Vector[FullOpening] = Vector(")

    for line in itertools.islice(open(argv[1]), 1, None):
        cols = line.split("\t")
        print(f"""new FullOpening("{cols[0]}", "{cols[1]}", "{cols[2]}"),""")

    print("  )")
    print("}")


if __name__ == "__main__":
    main(sys.argv)



from typing import Optional,  Callable
import subprocess
import platform
import shutil
import glob

def installLeipFont(notify : Optional[Callable[[str], None]] = None) -> None:
 import sys, os, os.path
 sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
 import MzChess
 
 if MzChess.useQt5():
  from PyQt5 import QtGui
  families = QtGui.QFontDatabase().families()
 else:
  from PyQt6 import QtGui
  families = QtGui.QFontDatabase.families()
 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 leipfontFile = os.path.join(fileDirectory, 'pieces', 'LEIPFONT.ttf')
 if platform.system() == 'Linux_':
  fontFileList = map(lambda x : x.split(':')[0], subprocess.getstatusoutput('fc-list')[1].split('\n'))
  hasLEIPFONT = False
  for fileName in fontFileList:
   if os.path.basename(leipfontFile) in fileName:
    hasLEIPFONT = True
  if not hasLEIPFONT:
   fontDir = os.path.join(os.path.expanduser('~'), '.local', 'share', 'fonts')
   if not os.path.isdir(fontDir):
    os.mkdir(fontDir)
   for file in glob.glob(leipfontFile):
    shutil.copyfile(file, os.path.join(fontDir, file))
   os.system('fc-cache -f -v')
   if notify is not None:
    notify('Chess font {} added'.format(leipfontFile))
  elif notify is not None:
   notify('Chess font {} already installed'.format(leipfontFile))
 else:
  if 'Chess Leipzig' not in families:
   assert QtGui.QFontDatabase.addApplicationFont(leipfontFile) != -1, "installLeipFont: Could not add 'Chess Leipzig' font"
   if notify is not None:
    notify('Chess font {} added'.format(leipfontFile))
  elif notify is not None:
   notify('Chess font {} already installed'.format(leipfontFile))



'''
Game Header Editor
=====================

The game header editor is a 2-column table which assigns to values to header elements.
The `PGN`_ standard lists the supported header elements. 
The header elements of 7-tag roster, 
i.e. *Event*, *Site*, *Date*, *Round*, *White*, *Black* and *Result*, are mandatory.

The *Game/Select Header Elements ...* menu entry opens a dialog to add/remove header elements.

|HeaderEditor|

Depending on the type, the values are edited using a *text*, *date* and *time* editors.
To avoid inconsistent header data, the following header elements are readonly:

* *Result*, *Annotator*, *PlyCount*
* any kind of opening information, i.e. *Opening*, *Variation*, *SubVariation*, *ECO*, *NIC*


.. |HeaderEditor| image:: headerEditor.png
  :width: 800
  :alt: Header Editor
.. _PGN: https://github.com/fsmosca/PGN-Standard
'''
from typing import Optional, List, Tuple, Any
import copy
import datetime
from enum import IntEnum, unique
import sys,  os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtWidgets, QtCore
else:
 from PyQt6 import QtWidgets, QtCore

import chess, chess.pgn

@unique
class KeyType(IntEnum):
 STR = 0
 INT = 1
 LABEL = 2
 DATE = 3
 TIME = 4
 EVENT = 5
 SITE = 6
 PLAYER = 7
 TITLE = 8

class GameHeaderView(QtWidgets.QTableWidget):
 '''Game Header Editor object
 '''
 stdKey2ValueTypeDict = {
   # seven tag roster
   "Event" : ("?", KeyType.EVENT),
   "Site" : ("?", KeyType.SITE),
   "Date" : ("????.??.??", KeyType.DATE),
   "Round" : ("?", KeyType.STR),
   "White" : ("?", KeyType.PLAYER),
   "Black" : ("?", KeyType.PLAYER),
   "Result" : ("*", KeyType.LABEL), 
   # player related information 
   "WhiteTitle" : ("", KeyType.TITLE), 
   "BlackTitle" : ("", KeyType.TITLE), 
   "WhiteElo" : ("", KeyType.INT), 
   "BlackElo" : ("", KeyType.INT), 
   "WhiteUSCF" : ("", KeyType.INT), 
   "BlackUSCF" : ("", KeyType.INT), 
   "WhiteNA" : ("", KeyType.STR), 
   "BlackNA" : ("", KeyType.STR), 
   # event related information 
   "EventDate" : ("", KeyType.DATE),
   "EventSponsor" : ("", KeyType.STR),
   "Section" : ("", KeyType.STR),
   "Stage" : ("", KeyType.STR),
   "Board" : ("", KeyType.INT),
   # opening information 
   "Opening" : ("", KeyType.LABEL), 
   "Variation" : ("", KeyType.LABEL), 
   "SubVariation" : ("", KeyType.LABEL), 
   "ECO" : ("", KeyType.LABEL), 
   "NIC" : ("", KeyType.LABEL), 
   # time
   "Time" : ("??:??:??", KeyType.TIME),
   "UTCTime" : ("??:??:??", KeyType.TIME),
   "UTCDate" : ("????.??.??", KeyType.DATE),
   # time control
   "TimeContol" : ("?", KeyType.STR), 
   # alternative starting positions
   "FEN" : (chess.STARTING_FEN, KeyType.LABEL), 
   # game termination
   "Termination" : ("", KeyType.STR), 
   # miscellaneous
   "Annotator" : ("", KeyType.LABEL), 
   "PlyCount" : ("0", KeyType.LABEL)}
 gameResult = ["1-0", "0-1", "1/2-1/2", "*"]

 def __init__(self, parent = None) -> None:
  super(GameHeaderView, self).__init__(parent)
  self.notifyGameHeadersChangedSignal = None
  self.gameResultLabel = None
  self.eventList = ['?']
  self.siteList = ['?']
  self.playerList = ['?']
  
 def setup(self, notifyGameHeadersChangedSignal  : Optional[QtCore.pyqtSignal] = None, 
                 eventList : List[str] = list(), siteList : List[str] = list(), playerList : List[str] = list()) -> None:
  self.eventList = ['?'] + eventList
  self.siteList = ['?'] + siteList
  self.playerList = ['?'] + playerList
  self.notifyGameHeadersChangedSignal = notifyGameHeadersChangedSignal
  self.resetGame()
 
 @staticmethod
 def headerElements(withoutSevenTagRoster : bool = False) -> List[str]:
  '''Get game header elements

:param withoutSevenTagRoster: returns only the optional header elements
:returns: list of header elements
  '''
  if not withoutSevenTagRoster:
   hElements = GameHeaderView.stdKey2ValueTypeDict.keys()
  else:
   hElements = list()
   for el in GameHeaderView.stdKey2ValueTypeDict.keys():
    if el not in chess.pgn.Headers():
     hElements.append(el)
  return hElements

 def _createCompleteEdit(self, selList : List[str], value : Any) -> QtWidgets.QLineEdit:
  completer = QtWidgets.QCompleter(selList)
  completer.setCaseSensitivity(QtCore.Qt.CaseSensitivity.CaseInsensitive)
  completer.setFilterMode(QtCore.Qt.MatchFlag.MatchContains)
  completer.setCompletionMode(QtWidgets.QCompleter.CompletionMode.InlineCompletion)
  item = QtWidgets.QLineEdit(str(value))
  item.setCompleter(completer)
  item.editingFinished.connect(self.on_editingFinished)
  return item

 def _showTable(self, keyValueTypeList : List[Tuple[str, str, KeyType]]) -> None:
  self.horizontalHeader().hide()
  self.verticalHeader().hide()
  self.setRowCount(len(keyValueTypeList))
  self.setColumnCount(2)
  for row, self.defaultKeyValueType in enumerate(keyValueTypeList):
   key, value, type = self.defaultKeyValueType
   item = QtWidgets.QLabel(str(key))
   self.setCellWidget(row, 0, item)
  test = QtWidgets.QLabel()
  zeroWidth = test.fontMetrics().size(QtCore.Qt.TextFlag.TextSingleLine, '0').width()
  self.setColumnWidth(0, 20*zeroWidth)
  
  self.itemList = list()
  for row, self.defaultKeyValueType in enumerate(keyValueTypeList):
   key, value, type = self.defaultKeyValueType
   if type == KeyType.STR:
    item = QtWidgets.QLineEdit(str(value))
    item.editingFinished.connect(self.on_editingFinished)
   elif type == KeyType.LABEL:
    item = QtWidgets.QLabel(str(value))
    item.setStyleSheet("background-color : white; color : gray;")
    if key == 'PlyCount':
     self.plyCountLabel = item
    elif key == 'Result':
     self.gameResultLabel = item
   elif type == KeyType.DATE:
    try:
     datetime.date.fromisoformat(value.replace('.','-'))
     date = QtCore.QDate.fromString(value, "yyyy.MM.dd")
    except:
     date = QtCore.QDate.currentDate()
     self.gameHeaders[key] = date.toString('yyyy.MM.dd')
    item = QtWidgets.QDateEdit(date)
    item.dateChanged.connect(self.on_dateChanged)
   elif type == KeyType.TIME:
    try:
     _time = QtCore.QTime.fromString(value, "hh:mm:ss")
    except:
     _time = QtCore.QTime.currentTime()
    item = QtWidgets.QTimeEdit(_time)
    item.timeChanged.connect(self.on_timeChanged)
   elif type == KeyType.INT:
    item = QtWidgets.QSpinBox()
    item.setMaximum(4000)
    item.setSpecialValueText('-')
    item.valueChanged.connect(self.on_spinBox_changed)
   elif type == KeyType.EVENT:
    item = self._createCompleteEdit(self.eventList, value)
   elif type == KeyType.SITE:
    item = self._createCompleteEdit(self.siteList, value)
   elif type == KeyType.PLAYER:
    item = self._createCompleteEdit(self.playerList, value)
   else:
    continue
   
   self.itemList.append(item)
   self.setCellWidget(row, 1, item)
  self.update()
  
 def _getHeaders(self) -> chess.pgn.Headers:
  return self.gameHeaders

 def resetGame(self) -> None:
  '''Resets editor to standard header
  '''
  self.setGame(chess.pgn.Game())

 def setPlyCount(self, newCount : int) -> None:
  '''Sets the 'PlyCount' header element, if selected 

:param newCount: actual number of halfmoves
  '''
  if self.plyCountLabel is not None:
   self.gameHeaders['PlyCount'] = newCount
   self.plyCountLabel.setText(str(newCount))

 def setGameResult(self, result : str) -> None:
  '''Sets the 'Result' header element

:param result: one out of *1-0*, *0-1*, *1/2-1/2*, *\**
  '''
  if result not in self.gameResult:
   result = self.gameResult[3]
  self.gameHeaders['Result'] = result
  self.gameResultLabel.setText(result)

 def setGame(self, game : chess.pgn.Game) -> None:
  '''Edit the header of a game. It corrects the following header elements
  
  * *Result*
  * *FEN*, if available
  * *PlyCount*, if available

:param game: game to be edited
  '''
  self.clear()
  self.gameHeaders = copy.deepcopy(game.headers)
  for key, defValue in chess.pgn.Headers().items():
   if key not in self.gameHeaders:
    self.gameHeaders[key] = defValue
  if 'FEN' in self.gameHeaders:
   self.gameHeaders['FEN'] = game.board().fen()
  if self.gameHeaders['Result'] not in self.gameResult:
   value = self.gameResult[3]
  plyCount = 0
  gameNode = game
  while gameNode is not None:
   plyCount += 1
   gameNode = gameNode.next()
  if 'PlyCount' in self.gameHeaders:
   self.gameHeaders['PlyCount'] = str(plyCount)
  keyValueTypeList = list()
  for key, value in self.gameHeaders.items():
   if key in self.stdKey2ValueTypeDict:
    type = self.stdKey2ValueTypeDict[key][1]
   else:
    key = '??? {} ???'.format(key)
    type = KeyType.STR
   keyValueTypeList.append((key, value, type))
  self._showTable(keyValueTypeList)
 
 def _getKV(self) -> Tuple[int, str, QtWidgets.QLineEdit]:
  item = self.sender()
  row = self.itemList.index(item)
  key = self.cellWidget(row, 0).text()
  return key, self.cellWidget(row, 1)
  
 @QtCore.pyqtSlot()
 def on_editingFinished(self):
  key, valueItem = self._getKV()
  text = valueItem.text().strip(' \t\n')
  self.gameHeaders[key] = text
  self._emitHeader()
  
 @QtCore.pyqtSlot(QtCore.QDate)
 def on_dateChanged(self, date):
  key, _ = self._getKV()
  self.gameHeaders[key] = date.toString('yyyy.MM.dd')
  self._emitHeader()

 @QtCore.pyqtSlot(QtCore.QTime)
 def on_timeChanged(self, _time):
  key, _ = self._getKV()
  self.gameHeaders[key] = _time.toString('hh:mm:ss')
  self._emitHeader()

 @QtCore.pyqtSlot(int)
 def on_spinBox_changed(self, value):
  key, _ = self._getKV()
  if value != 0:
   self.gameHeaders[key] = str(value)
  else:
   self.gameHeaders[key] = ''
  self._emitHeader()
   
 def _emitHeader(self) -> None:
  for key, value in self.gameHeaders.items():
   if len(value) == 0:
    del self.gameHeaders[key]
  if self.notifyGameHeadersChangedSignal is not None:
   self.notifyGameHeadersChangedSignal.emit(self.gameHeaders)
  
if __name__ == "__main__":
 import io, sys
 from pgnParse import read_game

 if True:
  newData = """[Event "matein2"]
[Site "problem solved"]
[Date "????.??.??"]
[Round "?"]
[White "?"]
[Black "?"]
[Result "*"]
[Time "??:??:??"]
[FEN "1k6/Rp1K4/1P5P/8/P7/3pP3/1p1P4/8 w - - 0 1"]
  
1.h7 b1=Q! $20 {[% -4.80]} 2.h8=R# *"""
 else:
  ps = "C:/Users/Reinh/OneDrive/Dokumente/Schach/ps.pgn"
  with open(ps, mode = 'r',  encoding = 'utf-8') as f:
   newdata = f.read()
 
 pgn = io.StringIO(newData)
 game = read_game(pgn)
 app = QtWidgets.QApplication([])
 tbl = GameHeaderView()
 tbl.setup(
   eventList = ["Wöchentliche Schachmeisterschaften"], 
   siteList = ["München GER", "Internet"], 
   playerList = ["März, Reinhard", "Schlüter, Paul"])
 tbl.setGame(game)
 tbl.resize(360,240)
 tbl.show()
 sys.exit(app.exec_())



from typing import Callable, TextIO, List

import pytest

import os, os.path
from PyQt6 import QtCore

import MzChess
import chess

def runUciEngine(executable : str, fenList : List[str]) -> None:
 print('runUciEngine: {}\n fenList = {}'.format(executable, fenList))
 def waitForData(engine):
  for i in range(100):
   QtCore.QThread.msleep(engine.timeout_msec)
   QtCore.QCoreApplication.processEvents()
   if engine.isReady():
    return
   print('----> waiting {} * {} ms'.format(i, engine.timeout_msec))

 def printPlayResult(engine):
  playResult = engine.playResult
  rootBoard = chess.Board(engine.board.fen())
  print('{} : {}'.format('best move', playResult.move))
  if playResult.ponder is not None:
   print('{} : {}'.format('expected response', playResult.ponder))
  if isinstance(playResult.info, dict):
   infoList = [playResult.info]
  else:
   infoList = playResult.info
  for i, infoDict in enumerate(infoList):
   print('info #{}'.format(i+1))
   for key, value in infoDict.items(): 
    if key == 'pv':
     pvString = ''
     for n, moveList in enumerate(value):
      if not isinstance(moveList, list):
       moveList = [moveList]
      for move in moveList: 
       rootBoard.push(move)
       uciMove = move.uci()
       if rootBoard.is_checkmate():
        uciMove += '#'
       elif rootBoard.is_check():
        uciMove += '+'
       if engine.board.turn:
        if n == 0:
         pvString += '1.{}'.format(uciMove)
        elif n % 2 == 0:
         pvString += ' {}.{}'.format(n//2+1, uciMove)
        else:
         pvString += ' {}'.format(uciMove)
       else:
        if n == 0:
         pvString += '1...{}'.format(uciMove)
        elif n % 2 == 1:
         pvString += ' {}.{}'.format((n+1)//2+1, uciMove)
        else:
         pvString += ' {}'.format(uciMove)
       print('{} : {}'.format(key, pvString))
    elif value is not None: 
     print('{} : {}'.format(key, value))
 
 # =====================================================================

 engine = MzChess.ChessEngine((executable,  dict()) , limit = chess.engine.Limit(depth = 15),  log = print)
 waitForData(engine)
 for item in sorted(engine.idDict):
  print(' {} : {}'.format(item, engine.idDict[item]))
 print('Options:')
 for opt in sorted(engine.optionsDict):
  print(' {} : {}'.format(opt, engine.optionsDict[opt]))
 for fen in fenList:
  engine.uciNewGame(fen = fen)
  print('------------------------- startPlay -------------------------')
  engine.startPlay()
  waitForData(engine)
  printPlayResult(engine)
  print('----------------------- startAnalysis -----------------------')
  engine.startAnalysis(multiPV = 1)
  waitForData(engine)
  printPlayResult(engine)
 engine.kill()
 pytest.helpers.exitApp()

def test_uciEngine(qtbot, pytestconfig):
 fen = pytestconfig.getoption("--FEN")
 if fen is not None:
  fenList = [fen]
 else:
  fenList = [
  "3rr3/2p3p1/4N3/3b4/1p3P2/2PBB3/kPQ3PP/3R2K1 w - - 0 1", 
  "rn3r1k/pp4pp/2p1Q2N/q2np3/4N2P/2PP4/PP3P2/R3K2R w KQ - 0 1", 
  "r2rkn2/1R1N2p1/2p1B2p/4p1PP/p1P2b2/5P2/P3K3/1R6 w - - 0 1", 
  "2r4k/p2R4/1p2P3/5p2/3PbbpP/BP6/P1r5/4Q1K1 b - - 0 1"
  ]
 uciEngine = pytestconfig.getoption("--uciEngine")
 if uciEngine is None:
  pytest.skip('uciEngine not specified')
  return
 if not os.path.isabs(uciEngine): 
  home = pytestconfig.getoption("--home")
  uciEngine = os.path.join(home, uciEngine) 
 if not os.path.exists(uciEngine):
  raise IOError('Engine {} not found'.format(uciEngine))
  return
 runUciEngine(uciEngine, fenList)
 return




# Try using setuptools first, if it's installed
try:
 from setuptools import setup
except:
 print('Install the setuptools module (pip3 install setuptools)')
 quit()

if __name__ == '__main__':
    setup()
    quit()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module implementing the post install tasks not done by 'pip install'.
"""

import sys
import os, os.path
import platform
import shutil

def postInstall():
 osName = platform.system()
 home = os.path.expanduser('~')
 scripts = ['mzChess',  'buildFen']
 scriptIcons = ['schach.png',  'schach.png']
   
 if osName == 'Windows':
  import ctypes
  if ctypes.windll.shell32.IsUserAnAdmin() != 1:
   raise PermissionError('Admin privileges required to run this script.')
  scriptsDir = os.path.join(os.path.dirname(sys.executable), 'Scripts')
  desktopDir = os.path.join(home, 'Desktop')
  for script in scripts:
   src = os.path.join(scriptsDir,'{}.exe'.format(script))
   scriptTgt = os.path.join(scriptsDir, script)
   tgt = os.path.join(desktopDir, script)
   if os.path.exists(scriptTgt):
    os.remove(scriptTgt)
   os.symlink(src, scriptTgt)
   if os.path.exists(tgt):
    os.remove(tgt)
   shutil.move(scriptTgt, tgt)
 elif osName == 'Linux':
  import MzChess
  packageDir = os.path.dirname(MzChess.__file__)
  desktopDir = os.path.join(home, '.local', 'share', 'applications')
  iconsDir = os.path.join(home, '.local', 'share', 'icons')
  hasDeskopFileInstall = True
  for script, icon in zip(scripts, scriptIcons):
   iconPath = os.path.join(packageDir, icon)
   shutil.copy(iconPath, iconsDir, follow_symlinks = False)
   scriptPath = shutil.which(script)
   src = os.path.join(packageDir, '{}.desktop'.format(script))
   tgt = os.path.join(home, '{}.desktop'.format(script))
   with open(src, "r", encoding="utf-8") as f:
    text = f.read()
   text = text.format(scriptPath = scriptPath)
   with open(tgt, "w", encoding="utf-8") as f:
    f.write(text)
   if os.system('desktop-file-install --delete-original --dir={} {}'.format(desktopDir, tgt)) == -1:
    os.chmod(tgt, 0o644)
    shutil.move(tgt, desktopDir)
    hasDeskopFileInstall = False
  if not hasDeskopFileInstall:
   raise OSError('desktop-file-install failed, .desktop files manually moved to {}'.format(desktopDir))
 else:
  raise OSError('OS "{}" is not supported'.format(osName))

if __name__ == "__main__":
 postInstall()




'''Annotation support

.. _fiekas.eco: https://github.com/niklasf/eco
'''

from typing import Callable, List, Union, Optional, Tuple
import sys
import os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5():
 from PyQt5 import QtCore
else:
 from PyQt6 import QtCore

import chess, chess.pgn
from chessengine import ChessEngine, PGNEval_REGEX

class Annotator():
 '''A Annotator class applying 

:param name: name of the annotation engine
:param notifyFunction: print-like function used for notification
 '''

 def __init__(self, name : str,  notifyFunction : Optional[Callable[[str], None]] = None) -> None:
  self.name = name
  self.notifyFunction = notifyFunction
  self.posScNagAV : List[Tuple[float, int, bool]] = list()
  self.negScNagAV : List[Tuple[float, int, bool]] = list()
 
 def _setClass(self, nagCode : int, score : float, addVariant : bool):
  if score > 0:
   newList = [(score, nagCode, addVariant)]
   for scNagAv in self.posScNagAV:
    if scNagAv[1] != nagCode:
     newList.append(scNagAv)
   self.posScNagAV = sorted(newList, reverse = True, key = lambda el : el[1])
  else:
   newList = [(score, nagCode, addVariant)]
   for scNagAv in self.negScNagAV:
    if scNagAv[1] != nagCode:
     newList.append(scNagAv)
   self.negScNagAV = sorted(newList, reverse = False, key = lambda el : el[1])
  
 # ---------------------------------------------------------------

 def setBlunder(self, score : float, addVariant : bool = True) -> None:
  '''Sets the condition for a blunder move (NAG: $4)
  
:param score: score limit (score < 0 expected)
:param addVariant: add a variant in below score
  '''
  if score > 0:
   score = -score
  self._setClass(chess.pgn.NAG_BLUNDER, score, addVariant)

 def setDubiousMove(self, score : float, addVariant : bool = False) -> None:
  '''Sets the condition for a dubious move (NAG: $6)
  
:param score: score limit (score < 0 expected)
:param addVariant: add a variant in below score
  '''
  if score > 0:
   score = -score
  self._setClass(chess.pgn.NAG_DUBIOUS_MOVE, score, addVariant)

 def setPoorMove(self, score : float, addVariant : bool = False) -> None:
  '''Sets the condition for a poor move (NAG: $2)
  
:param score: score limit (score < 0 expected)
:param addVariant: add a variant in below score
  '''
  if score > 0:
   score = -score
  self._setClass(chess.pgn.NAG_MISTAKE, score, addVariant)

 def setBrillantMove(self, score : float, addVariant : bool = False) -> None:
  '''Sets the condition for a brilliant move (NAG: $3)
  
:param score: score limit (score > 0 expected)
:param addVariant: add a variant in below score
  '''
  if score < 0:
   score = -score
  self._setClass(chess.pgn.NAG_BRILLIANT_MOVE, score, addVariant)

 def setSpeculativeMove(self, score : float, addVariant : bool = False) -> None:
  '''Sets the condition for a speculative move (NAG: $5)
  
:param score: score limit (score > 0 expected)
:param addVariant: add a variant in below score
  '''
  if score < 0:
   score = -score
  self._setClass(chess.pgn.NAG_SPECULATIVE_MOVE, score, addVariant)

 def setGoodMove(self, score : float, addVariant : bool = False) -> None:
  '''Sets the condition for a good move (NAG: $1)
  
:param score: score limit (score > 0 expected)
:param addVariant: add a variant in below score
  '''
  if score < 0:
   score = -score
  self._setClass(chess.pgn.NAG_GOOD_MOVE, score, addVariant)
  
 # ---------------------------------------------------------------

 def _nagAv(self, whiteScore : str, lastWhiteScore : str, turn):
  if whiteScore is not None and len(whiteScore) > 0:
   try: 
    score = float(whiteScore)
   except:
    score = float(whiteScore[1:])
   try: 
    lastScore = float(lastWhiteScore)
   except:
    lastScore = float(lastWhiteScore[1:])
   if not turn:
    score = -score
    lastScore = -lastScore
   if score > 0 and len(self.posScNagAV) > 0:
    for sc, nag, av in self.posScNagAV:
     if lastScore - score > sc:
      return ([nag], av)
   elif score < 0 and len(self.negScNagAV) > 0:
    for sc, nag, av in self.negScNagAV:
     if lastScore - score < sc:
      return ([nag], av)
  return (list(), False)
   
 def _replaceScore(self, comment : str, newWhiteScore : float):
  while True:
   match = PGNEval_REGEX.search(comment)
   if match is None:
    break
   comment = comment.replace(match.group(0),'')
  comment = '[%eval {}] {}'.format(newWhiteScore, comment)
  return comment 

 @staticmethod
 def remove(game : chess.pgn.Game, comments : bool = False, variants : bool = False) -> chess.pgn.Game:
  '''Remove certain items from a game
  
:param game: game to process
:param comments: If True, remove all comments
:param variants: If True, remove all variants
  '''

  gameNode = game
  while gameNode is not None:
   if comments:
    gameNode.comment = ''
   if variants and len(gameNode.variations) > 0:
    gameNode.variations = [gameNode.variations[0]]
   gameNode = gameNode.next()
  return game
  
 def apply(self, 
                game : Union[chess.pgn.Game, chess.pgn.GameNode] = chess.pgn.Game(), 
                scoreListList : List[List[float]] = list(), 
                pvListList : Optional[List[List[List[chess.Move]]]] = None, 
                forceHints : bool = False) -> None:
  '''Apply the results of AnnotateEngine.run to a game 
  
:param game: game or gameNode where annotation starts (required)
:param scoreListList: for each move a list of scores for each variant, see AnnotateEngine.scoreListList
:param pvListList: for each move a list of lists of moves for each variant, see AnnotateEngine.pvListList
:param forceHints: force the creation of variants independent of the setXX definitions
  '''
  
  assert len(scoreListList) > 0, 'Empty scoreListList detected'
  if pvListList is not None:
   if len(pvListList) == 0:
    pvListList = None
   else:
    assert len(scoreListList) == len(pvListList), 'len(scoreListList) == len(pvListList) required'
  if isinstance(game, chess.pgn.Game):
   gameNode = game.next()
   if len(self.name) > 0:
    game.headers['Annotator'] = self.name
  else:
   gameNode = game
  lastWsc = 0
  for plyID, scoreList in enumerate(scoreListList):
   if gameNode is None:
    break
   if pvListList is not None:
    pvList = pvListList[plyID]
   else:
    pvList = None
   wsc = scoreList[0]
   nag, av = self._nagAv(wsc, lastWsc, gameNode.turn())
   lastWsc = wsc
   gameNode.nags = nag
   gameNode.comment = self._replaceScore(gameNode.comment, wsc)
   addHints = (av or forceHints) and pvList is not None
   if self.notifyFunction is not None:
    self.notifyFunction('{}. {}: score = {}, nags = {}'.format(plyID, gameNode.move, wsc, nag))
   if addHints:
    for n, score in enumerate(scoreList):
     if self.notifyFunction is not None:
      self.notifyFunction('--> {}'.format(pvList[n]))
     gameNode.parent.add_line(pvList[n], starting_comment = '<-[%eval {}]'.format(score))
   gameNode = gameNode.next()
  return game

class AnnotateEngine(QtCore.QObject):
 '''A wrapper class collecting score and variant (pv) data from an engine 

:param notifyFunction: print-like function used for notification
 '''
 def __init__(self, 
                    notifyFunction : Optional[Callable[[str], None]] = None, 
                    parent : Optional[QtCore.QObject] = None) -> None:
  super(AnnotateEngine, self).__init__(parent)
  self.notifyFunction = notifyFunction

 def setup(self, 
                engine : ChessEngine, 
                hintPLYs : int = 0, 
                multiPV : int = 1) -> None:
  '''Setup for operation

:param engine: engine used for annotation
:param hintPLYs: number of half moves (plys) in variants. Suppress hints by setting hintPLYs == 0 
:param multiPV: number of variants
  '''
  assert hintPLYs >= 0
  assert multiPV > 0
  self.halfMoveID = 0
  self.engine = engine
  self.engine.bestMoveScoreSignal.connect(self._bestMoveScoreAvailable)
  self.hintPLYs = hintPLYs
  self.multiPV = multiPV

 @QtCore.pyqtSlot(chess.Move, str)
 def _bestMoveScoreAvailable(self, move : chess.Move, score : str):
  if len(score) == 0:
   score = None
  if self.notifyFunction is not None:
   if self.gameNode.move is not None:
    san = self.gameNode.san()
   else:
    san = None
   if not self.gameNode.turn():
    moveText = '{}. {}'.format(self.halfMoveID//2 + 1, san)
   else:
    moveText = '... {}'.format(san)
   self.notifyFunction('{}: score = {}'.format(moveText, score))
  self.halfMoveID += 1
  scoreList = list()
  pvList = list()
  if not isinstance(self.engine.playResult.info, list):
   self.engine.playResult.info = [self.engine.playResult.info]
  for hintID, info in enumerate(self.engine.playResult.info):
   if self.hintPLYs > 0:
    if 'pv' in info:
     pvList.append(info['pv'][:self.hintPLYs])
    else:
     pvList.append([])
   scoreList.append(self.engine.getScore(hintID = hintID))
  self.scoreListList.append(scoreList)
  if self.hintPLYs > 0:
   self.pvListList.append(pvList)
  if (self.numberOfPlys is not None and len(self.scoreListList) >= self.numberOfPlys) or not self._startNext():
   return
 
 def _startNext(self, isNew : bool = False) -> bool:
  if not isNew:
   self.gameNode = self.gameNode.next()
  if self.gameNode is None:
   return False
  self.engine.uciNewGame(fen = self.gameNode.board().fen())
  return self.engine.startAnalysis(multiPV = self.multiPV)
  
 def run(self, game : Union[chess.pgn.Game, chess.pgn.GameNode], numberOfPlys : Optional[int] = None) -> bool:
  '''Runs the engine for a whole game or a gameNode

:param game: game or gameNode
:param numberOfPlys: number of half moves to analyse, ``None`` means analysis of the rest of the game
:returns: True, if successful
  '''
  if isinstance(game, chess.pgn.Game):
   self.gameNode = game.next()
  else:
   self.gameNode = game
  self.numberOfPlys = numberOfPlys
  self.scoreListList = list()
  if self.hintPLYs > 0:
   self.pvListList = list()
  else:
   self.pvListList = None
  if not self._startNext(isNew = True):
   return False
  while (self.numberOfPlys is None or len(self.scoreListList) < self.numberOfPlys) and self.gameNode is not None:
   QtCore.QCoreApplication.processEvents()
  return True

if __name__ == "__main__":
 import os, sys
 import pickle
 import argparse
 import configparser
 import chess.pgn
 import configureEngine
 from pgnParse import read_game

 app = QtCore.QCoreApplication(sys.argv)

 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 os.chdir(fileDirectory)
 parser = argparse.ArgumentParser(description='PGN Parser: test')
 
 parser.add_argument("pgnFile", help = "PGN-File")
 parser.add_argument("--encoding",  metavar = 'encoding',  default = 'u', 
   choices=['u', 'utf-8-sig', 'i', 'iso-8859-1', 'a', 'ascii'], 
   help="target of parsing (g - games, h - headers, s - skip every second game, b - board, l - lexical analysis only)")
 parser.add_argument("--gameID", metavar = 'gameID', type = int, default=0, help="ID of the game to be used")
 parser.add_argument("--plyID", metavar = 'plyID', type = int, default=0, help="ID of the game to be used")
 parser.add_argument("--multiPV", metavar = 'multiPV', type = int, default=1, help="ID of the game to be used")
 parser.add_argument("--engine", metavar = 'engine', type = str, help="ID of the game to be used")
 parser.add_argument("-debug", action = 'store_true', default = False, help = "Enable Debugging")

 args = parser.parse_args()
 assert args.pgnFile is not None, 'pgnFile is required'
 encoding = args.encoding[0]
 if encoding == 'u':
  encoding = 'utf-8-sig'
 elif encoding == 'i':
  encoding = 'iso-8859-1'
 elif encoding == 'a':
  encoding = 'ascii'

 settingsFile = os.path.join(fileDirectory, 'settings.ini')
 settings = configparser.ConfigParser(delimiters=['='], allow_no_value=True)
 settings.optionxform = str
 settings.read(settingsFile, encoding = 'utf-8')
 engineDict = configureEngine.loadEngineSettings(settings)
 if args.engine is None:
  selectedEngine = settings['Menu/Engine']['selectedEngine']
 else:
  selectedEngine = args.engine
 assert selectedEngine in engineDict, 'Unexpected engine {} (must be out of {})'.format(selectedEngine,  engineDict)
 executable = engineDict[selectedEngine]

 engine = ChessEngine(executable, limit = chess.engine.Limit(depth = 15),  log = False)
 
 _,  ext = os.path.splitext(args.pgnFile)
 assert ext in ['.ppgn', '.pgn'], 'Unexpected file type {} of {}'.format(ext, args.pgnFile)
 
 if ext == '.ppgn':
  with open(args.pgnFile, mode = 'rb') as f:
   gameList = pickle.load(f)
   game = gameList[args.gameID]
 else:
  pgn = open(args.pgnFile, mode = 'r',  encoding = encoding)
  for n in range(args.gameID + 1):
   game = read_game(pgn)
 
 annotateEngine = AnnotateEngine(notifyFunction = print)
 annotateEngine.setup(engine, hintPLYs = 3, multiPV = args.multiPV)

 if args.plyID <= 0:
  rc = annotateEngine.run(game)
  gameNode = game
  forceHints = False
 else:
  gameNode = game
  for n in range(args.plyID):
   gameNode = gameNode.next()
  rc = annotateEngine.run(gameNode.parent, numberOfPlys = 2)
  forceHints = True
  
 if rc:
  annotator = Annotator(selectedEngine, notifyFunction = print)
  annotator.setBlunder(1.0, addVariant = True)
  annotator.apply(game = gameNode, 
                        scoreListList = annotateEngine.scoreListList, 
                        pvListList = annotateEngine.pvListList, 
                        forceHints = forceHints)
  exporter = chess.pgn.StringExporter(headers=True, variations=True, comments=True)
  pgnString = game.accept(exporter)
  print(pgnString)
  base, _ = os.path.splitext(args.pgnFile)
  with open('{}_new.pgn'.format(base), "w", encoding = 'iso-8859-1') as f:
   f.write(pgnString)
 
 app.exec()
 print('completed')
